// ColorTable.java, v1.02, 13 December 2010
/****************************************************
**  ColorTable
**  Copyright (C) 2010  Rich Franzen
**
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see
**  <http://www.gnu.org/licenses/gpl-3.0.html>.
****************************************************/
package us.r0k.sihwheel;

import java.awt.*;
import java.awt.event.MouseEvent;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.table.*;
import java.util.ArrayList;

// add cell coloring and header toolTips to JTable
public class ColorTable extends JTable
{
    private ArrayList<String>	columnTips;

    public ColorTable(TableModel tm) {
	super(tm);
	columnTips = new ArrayList<String>();
    }
    public ColorTable(int numRows, int numCols) {
	super(numRows, numCols);
	columnTips = new ArrayList<String>();
    }
    public ColorTable(Object[][] rowData, Object[] columnNames) {
	super(rowData, columnNames);
	columnTips = new ArrayList<String>();
    }

    // addColumn() doesn't need to add to columnTips because of the
    /// way setColunmnToolTips(), setColumnToolTip() and getToolTipText()
    /// handle insufficient size of columnTips.

    // I verified that removeColumn() doesn't need to have
    /// a columnTip removed.  Original column indices are
    /// preserved even when a beginning or middle column are removed.

    TableColumn getTableColumn(int column) {
	TableColumn tc = null;
	if (column < 0)  return(tc);
	if (column >= getColumnCount())  return(tc);
	tc = getColumnModel().getColumn(column);
	return(tc);
    }

    // call if want normal text column aligned
    public void setColumnAlignment(int column, int hzAlignment) {
	TableColumn tc = getTableColumn(column);
	if (tc == null) {
	    System.err.println("illegal table column: " + column);
	    return;
	}
	tc.setCellRenderer(new PositionRenderer(hzAlignment));
    }

    // call if want a column to be in color and aligned
    public void setColumnToColor(int column, int hzAlignment) {
	TableColumn tc = getTableColumn(column);
	if (tc == null) {
	    System.err.println("illegal table column: " + column);
	    return;
	}
	tc.setCellRenderer(new ColoredBackgroundRenderer(hzAlignment));
	tc.setCellEditor(new ColoredBackgroundEditor());
    }

    // return white or black based on the supplied background
    public Color getVisibleForeground(Color bg) {
	Color fg = Color.white;
	if (bg != null) {
	    float lumi = bg.getRed()   * .212671f +
	    		 bg.getGreen() * .715160f +
	    		 bg.getBlue()  * .072169f;
	    if (lumi > 155)  fg = Color.black;
	}
	return(fg);
    }

    public void setColumnToolTips(String tips[]) {
	columnTips = new ArrayList<String>();
	for (int i = 0; i < tips.length; i++)  columnTips.add(tips[i]);
    }
    public void setColumnToolTip(int column, String tip) {
	if ((column >= 0) && (column < getColumnCount())) {
	    columnTips.ensureCapacity(column + 1);
	    for (int i = columnTips.size(); i <= column; i++)
		columnTips.add(null);
	    columnTips.set(column, tip);
	}
    }

    //Implement table header tool tips.
    protected JTableHeader createDefaultTableHeader() {
	return new JTableHeader(columnModel) {
	    public String getToolTipText(MouseEvent e) {
		java.awt.Point p = e.getPoint();
		int index = columnModel.getColumnIndexAtX(p.x);
		int realIndex = columnModel.getColumn(index).getModelIndex();
		String out = null;
		if (realIndex < columnTips.size())
		    out = columnTips.get(realIndex);
		return out;
	    }
	};
    }

    class ColoredBackgroundRenderer extends JLabel
				    implements TableCellRenderer
    {
	int	hzAlignment = SwingConstants.LEFT;

	ColoredBackgroundRenderer(int hzAlignment)
	{
	    if ((hzAlignment == SwingConstants.LEFT)    ||
		(hzAlignment == SwingConstants.CENTER)  ||
		(hzAlignment == SwingConstants.RIGHT)   ||
		(hzAlignment == SwingConstants.LEADING) ||
		(hzAlignment == SwingConstants.TRAILING))
		this.hzAlignment = hzAlignment;
	    setOpaque(true);  // MUST do this for background to show up.
	}

	public Component getTableCellRendererComponent(
	    JTable table, Object label,
	    boolean isSelected, boolean hasFocus,
	    int row, int column)
	{
	    ColorBGText ct = (ColorBGText)label;

	    if (isSelected) {
		setBackground(table.getSelectionBackground());
		setForeground(table.getSelectionForeground());
	    }
	      else {
		setBackground(ct.color);
		setForeground(getVisibleForeground(ct.color));
	      }

	    if (hasFocus) {
		Border border = null;
		if (isSelected)
		    border = UIManager.getBorder("Table.focusSelectedCellHighlightBorder");
		if (border == null)
		    border = UIManager.getBorder("Table.focusCellHighlightBorder");
	    }
	      else
		setBorder(new EmptyBorder(1, 1, 1, 1));

	    setText(ct.text);
	    setHorizontalAlignment(hzAlignment);
	    return(this);
	}
    }

    // modified: http://www.exampledepot.com/egs/javax.swing.table/CustEdit.html
    class ColoredBackgroundEditor extends AbstractCellEditor
				  implements TableCellEditor
    {
	// This is the component that will handle the editing of the cell value
	private JTextField	jtf = new JTextField();
	private ColorBGText	colorBGText;

	// This method is called when a cell value is edited by the user.
	public Component getTableCellEditorComponent(
	    JTable table, Object value,
	    boolean isSelected, int rowIndex, int vColIndex)
	{
	    // 'value' is in the cell located at (rowIndex, vColIndex)
	    colorBGText = (ColorBGText)value;  // preserve entire value

	    if (isSelected) {
		// cell (and perhaps other cells) are selected
	    }

	    // Configure the component with the specified value
	    jtf.setText(colorBGText.text);

	    // Return the configured component
	    return (Component)jtf;
	}

	// This method is called when editing is completed.
	// It must return the new value to be stored in the cell.
	public Object getCellEditorValue() {
	    colorBGText.text = jtf.getText();
	    return colorBGText;  // with original color still attached
	}
    }

    class PositionRenderer extends JLabel
			   implements TableCellRenderer
    {
	int	hzAlignment = SwingConstants.LEFT;

	PositionRenderer(int hzAlignment) {
	    if ((hzAlignment == SwingConstants.LEFT)    ||
		(hzAlignment == SwingConstants.CENTER)  ||
		(hzAlignment == SwingConstants.RIGHT)   ||
		(hzAlignment == SwingConstants.LEADING) ||
		(hzAlignment == SwingConstants.TRAILING))
		this.hzAlignment = hzAlignment;
	    setOpaque(true);  // MUST do this for background to show up.
	}

	public Component getTableCellRendererComponent(
	    JTable table, Object label,
	    boolean isSelected, boolean hasFocus,
	    int row, int column)
	{
	    if (isSelected) {
		setBackground(table.getSelectionBackground());
		setForeground(table.getSelectionForeground());
	    }
	      else {
		setBackground(table.getBackground());
		setForeground(table.getForeground());
	      }

	    if (hasFocus) {
		Border border = null;
		if (isSelected)
		    border = UIManager.getBorder("Table.focusSelectedCellHighlightBorder");
		if (border == null)
		    border = UIManager.getBorder("Table.focusCellHighlightBorder");
	    }
	      else
		setBorder(new EmptyBorder(1, 1, 1, 1));

	    setText((String)label);
	    setHorizontalAlignment(hzAlignment);
	    return(this);
	}
    }
}

// the object within each colored-background cell
class ColorBGText
{
    public String	text = "";
    // default to Motif default L&F background color
    public Color	color = new Color(174, 178, 195);

    ColorBGText(String text, Color color) {
	if (text  != null)  this.text  = text;
	if (color != null)  this.color = color;
    }

    @Override
    public String toString() {
	return(text);
    }
}
