// HelpBox.java, v1.14, 11 February 2013
/****************************************************
**  HelpBox
**  Copyright (C) 2010  Rich Franzen
**
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see
**  <http://www.gnu.org/licenses/gpl-3.0.html>.
****************************************************/
package us.r0k.sihwheel;

import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.io.IOException;
import javax.swing.*;
import javax.swing.table.*;
import javax.swing.text.DefaultEditorKit;
import java.text.DecimalFormat;
import java.net.MalformedURLException;
import java.net.URL;
import us.r0k.ntc.*;

/**
***   Provide two dialogs.  The first is an HTML dialog which would
*** typically be used to privide help or other non-interactive
*** information display.
***   The second sort of HelpBox is a specialized Color Log tuned
*** specifically to the needs of the SIHwheel.
**/
public class HelpBox extends JDialog
{
    private ColorTable          colorTable;
    private DefaultTableModel   dtm;
    private QEBeanTableModel    qErrors;
    private ColorTable          errorTable;
    private JLabel              jlEntries;
    private int                 entries;
    private static HelpBox      logBox;

    final static int    ROW_HEIGHT = 20;
    final static Color  bgc = new Color(23, 23, 32);
    final static String columnNames[] = {
        "Name", "Distance", "Hex", "Luminance",
        "RGB", "Complement", "HSB", "QE", "Mismatch"
    };
    final static String columnTips[] = {
        "color name, background = clicked color",
        "distance from clicked color to named color, background is named color",
        "hex value of clicked color, with matching background color",
        "luminance of clicked color, with background the greyscale equivalent",
        "red, green, and blue components",
        "hex value of complement color, with matching background",
        "hue, saturation, and brightness components",
        "quantization error between the RGB 24-bit color and the 16-bit SIH color",
        "mismatch between clicked color name and the color name found"
    };
    final static String statColumnNames[] = {
        "Type", "Count",
        "Red QE", "Green QE", "Blue QE", "Average QE",
        "Distance"
    };
    final static int HEX_QE     = 1;
    final static int NAME_QE    = 2;
    final static int MM_QE      = 3;
    final static int SEARCH_ERR = 4;

    // general window for display of HTML page
    HelpBox(Frame frame, String title, boolean modal,
        String pUrlS, Dimension pSize)
    {
        super(frame, title, modal);

        final String    urlS  = pUrlS;
        final Dimension size  = pSize;
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
                JEditorPane ep = new JEditorPane();
                ep.setEditable(false);
                try {
                    URL url = getClass().getResource(urlS);
                    ep.setPage(url);
                    JScrollPane eps = new JScrollPane(ep);
                    eps.setPreferredSize(size);
                    getContentPane().add(eps, BorderLayout.CENTER);
                } catch (IOException ioE) {
                    System.err.println("Unable to display help pane");
                    ioE.printStackTrace();
                }
                pack();
                setLocationRelativeTo(null);
                setVisible(true);
            }
        });
    }

    // window for log; instantiated once, in a runnable of SIHwheel
    HelpBox(Frame frame, String title, Dimension size)
    {
        super(frame, title, false);
        Container content = getContentPane();
        content.setLayout(new BoxLayout(content, BoxLayout.Y_AXIS));
        setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);
        setPreferredSize(size);
        setLocationRelativeTo(null);
        logBox = this;

        dtm = new DefaultTableModel(columnNames, 0) {
            @Override
            public boolean isCellEditable(int row, int col) {
                return(col > 7);  // allow editing on final column only
            }
        };
        colorTable = new ColorTable(dtm);
        colorTable.setRowHeight(ROW_HEIGHT);
        colorTable.setColumnToColor(0, SwingConstants.LEADING);
        colorTable.setColumnToColor(1, SwingConstants.TRAILING);
        colorTable.setColumnToColor(2, SwingConstants.CENTER);
        colorTable.setColumnToColor(3, SwingConstants.TRAILING);
        colorTable.setColumnToColor(4, SwingConstants.CENTER);
        colorTable.setColumnToColor(5, SwingConstants.CENTER);
        colorTable.setColumnToColor(6, SwingConstants.CENTER);
        colorTable.setColumnToColor(7, SwingConstants.LEADING);
        colorTable.setColumnToColor(8, SwingConstants.LEADING);
        colorTable.setColumnToolTips(columnTips);

        JScrollPane jts = new JScrollPane(colorTable);
        content.add(jts);
        JPanel textPane = new JPanel();
        textPane.setLayout(new BoxLayout(textPane, BoxLayout.Y_AXIS));
        textPane.setBackground(bgc);
        textPane.setPreferredSize(new Dimension(size.width, 170));
        textPane.setMaximumSize(new Dimension(2 * size.width, 170));

        JPanel jpT = new JPanel();
        jpT.setBackground(bgc);
        JButton help = new JButton("Help");
        help.setBackground(bgc.brighter());
        help.setForeground(Color.lightGray);
        help.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                Dimension size = new Dimension(700, 400);
                HelpBox hb = new HelpBox(
                    SIHwheel.ownerFrame, "Color Log Help", false,
                    "/ColorLogHelp.html", size);
            }
        });
        jpT.add(help);
        JLabel wm = new JLabel("WM");  // add some space
        wm.setForeground(bgc);
        jpT.add(wm);

        JLabel header = new JLabel(
            "<html><b>Accumulated Quantization Error</b></html>");
        header.setForeground(new Color(222, 150, 100));
        jpT.add(header);
        jlEntries = new JLabel(" for 0 entries");
        jlEntries.setForeground(Color.lightGray);
        jpT.add(jlEntries);
        jpT.setToolTipText("changing Color List resets all values");
        textPane.add(jpT);

        qErrors = new QEBeanTableModel(statColumnNames);
        qErrors.addError(new QEBean("Hex QE"));
        qErrors.addError(new QEBean("Name QE"));
        qErrors.addError(new QEBean("Mismatch QE"));
        qErrors.addError(new QEBean("Search Accuracy"));
        errorTable = new ColorTable(qErrors);
        errorTable.setBackground(bgc);
        errorTable.setForeground(Color.lightGray);
        errorTable.setRowHeight(20);
        // column 0 defaults to LEFT
        errorTable.setColumnAlignment(1, SwingConstants.CENTER);
        errorTable.setColumnAlignment(2, SwingConstants.RIGHT);
        errorTable.setColumnAlignment(3, SwingConstants.RIGHT);
        errorTable.setColumnAlignment(4, SwingConstants.RIGHT);
        errorTable.setColumnAlignment(5, SwingConstants.RIGHT);
        errorTable.setColumnAlignment(6, SwingConstants.CENTER);
        errorTable.setColumnToolTip(6, "(non-0 min, average, max)");
        JTableHeader jth = errorTable.getTableHeader();
        jth.setBackground(Color.black);
        jth.setForeground(Color.white);
        jth.setPreferredSize(new Dimension(jth.getWidth(), 22));
        textPane.add(jth);
        textPane.add(errorTable);

        JLabel copyHint = new JLabel(
            "to copy to clipboard, select one or more rows, then click Ctrl+C");
        copyHint.setForeground(Color.white);
        JPanel centerer = new JPanel();
        centerer.setBackground(bgc);
        centerer.add(copyHint);
        textPane.add(centerer);
        content.add(textPane);
        pack();

        float colorWidths[] = { .15f, .07f, .08f, .06f, .12f, .08f, .15f, .22f };
        setTableWidths(colorWidths, colorTable);
        float errorWidths[] = { .15f, .15f, .12f, .12f, .12f, .12f };
        setTableWidths(errorWidths, errorTable);
    }

    // add a new row of data
    /// values[0] = name OR CND name
    /// values[1] = quantized hex (derive rgb, hsb, and luminance) OR CND size
    /// values[2] = complement hex
    /// values[3] = mismatch (may be null or empty)
    /// values[4] = quantization error (may be null or empty)
    /// values[5] = exact hex
    // data is assumed to be valid
    public void newRow(String values[])
    {
        final String v[] = values;
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                DecimalFormat pct = new DecimalFormat("#0.0%");
                DecimalFormat f3p = new DecimalFormat("#0.000");
                int     rgb[]  = { 0, 0, 0 };
                int     rgbE[] = { 0, 0, 0 };
                int     rgbC[] = { 0, 0, 0 };
                float   hsb[]  = { 0, 0, 0 };
                int     luminance;
                ColorBGText row[] = new ColorBGText[9];
                for (int i = 0; i < 9; i++)
                    row[i] = new ColorBGText(null, null);

                // handle hue-breakdown report
                if (v.length == 1) {
                    row[7].text = v[0];  // by default widest column
                    dtm.addRow(row);
                    colorTable.setRowHeight(
                        dtm.getRowCount() - 1, ROW_HEIGHT * 7);
                    if (logBox.isVisible())  makeLastRowVisible();
                    return;
                }

                // handle display of color-name dictionary
                if (v.length == 2) {
                    row[0].text = v[0];
                    row[1].text = v[1];
                    dtm.addRow(row);
                    if (logBox.isVisible())  makeLastRowVisible();
                    return;
                }

                NTC.getRGB(v[1], rgb);
                NTC.getRGB(v[5], rgbE);
                NTC.getRGB(v[2], rgbC);
                Color.RGBtoHSB(rgb[0], rgb[1], rgb[2], hsb);
                luminance = (int)(rgb[0] * .212671 +
                    rgb[1] * .715160 + rgb[2] * .072169);
                Color quantized = new Color(rgb[0], rgb[1], rgb[2]);
                row[0].text  = v[0];
                row[0].color = quantized;
                row[1].text  = f3p.format(NTC.getDistance());
                row[1].color = new Color(rgbE[0], rgbE[1], rgbE[2]);
                row[2].text  = v[1];
                row[2].color = quantized;
                row[3].text  = Integer.toString(luminance);
                row[3].color = new Color(luminance, luminance, luminance);
                row[4].text  = "(" + rgb[0] + ", " + rgb[1] + ", " + rgb[2] + ")";
                row[4].color = quantized;
                row[5].text  = v[2];
                row[5].color = new Color(rgbC[0], rgbC[1], rgbC[2]);
                row[6].text = "(" + (int)(360 * hsb[0]) + ", " +
                         pct.format(hsb[1]) + ", " + (int)(hsb[2] * 255) + ")";
                row[6].color = quantized;
                row[7].text  = v[4];
                row[7].color = getColorAfterHash(v[4]);
                row[8].text  = v[3];
                row[8].color = getColorAfterHash(v[3]);

                dtm.addRow(row);
                if (logBox.isVisible())  makeLastRowVisible();
                if (dtm.getRowCount() > 2500) {
                    System.out.print("newRow(): removing first 500 rows ...");
                    for (int i = 0; i < 500; i++)  dtm.removeRow(0);
                    System.out.println(" complete.  2000 rows remain.");
                }
                colorTable.clearSelection();
            }
        });
    }

    public void makeLastRowVisible() {
        int rows = dtm.getRowCount();
        if (rows == 0)  return;
        Rectangle r = colorTable.getCellRect(rows - 1, 0, true);
        colorTable.scrollRectToVisible(r);
    }

    Color getColorAfterHash(String text) {
        Color color = new Color(174, 178, 195);  // default Motif background
        if (text == null)  return(color);
        int rgb[] = { 0, 0, 0 };
        int hash = text.lastIndexOf('#');
        if (hash < 0)  return(color);   // no # found

        String hex = text.substring(hash+1, hash+7);
        NTC.getRGB(hex, rgb);
        color = new Color(rgb[0], rgb[1], rgb[2]);
        return(color);
    }


    // update quantization error from external class
    public void updateQE(int which, String q, String a)
    {
        Color   quantized = Color.black;
        Color   actual = Color.black;
        int     rgb[] = { 0, 0, 0 };

        NTC.getRGB(q, rgb);
        quantized = new Color(rgb[0], rgb[1], rgb[2]);
        NTC.getRGB(a, rgb);
        actual = new Color(rgb[0], rgb[1], rgb[2]);
        if (quantized.equals(actual))  return;
        qErrors.getError(which - 1).updateData(quantized, actual);
        errorTable.clearSelection();
    }

    // update potential QE errors from external class
    public void updateQE(int which)
    { // count entries which theoretically could cause QE
        qErrors.getError(which - 1).updateEntries();
        entries += 1;
        jlEntries.setText(" for " + entries + " entries");
    }

    public void resetQE() {
        qErrors.resetAll();
        entries = 0;
        jlEntries.setText(" for 0 entries");
    }

    // call after the table has been instantiated and sized
    //// percent is array of tablewidths, count <= table width
    //// table is the table whose column widths are being changed
    static void setTableWidths(float percent[], JTable table)
    {
        TableColumn     tc;
        int             pixelsLeft, pixelsNow;
        int             column = 0;

        if ((percent == null) || (table == null)) {
            System.err.println("cannot set column widths; null paramater");
            return;
        }
        if (percent.length == 0)  return;

        // do sanity chack on array of floats
        float   sum = 0;
        for (int i = 0; i < percent.length; i++) {
            if (percent[i] <= 0) {
                sum = 2f;  // flag the problem
                break;
            }
            sum += percent[i];
        }
        if (sum > 1.0001f) {
            System.err.println("cannot set column widths; floats nonsensicle");
            return;
        }

        // the logic finally begins here
        int     overall = table.getWidth();
        int     columns = table.getColumnCount();

        pixelsLeft = overall;

        // handle first n rows, up to length of percent[]
        while (column < columns) {
            pixelsNow = (int)(percent[column] * overall);
            tc = table.getColumnModel().getColumn(column);
            tc.setPreferredWidth(pixelsNow);
            pixelsLeft -= pixelsNow;
            column += 1;
            if (column >= percent.length)  break;
        }
        if (column >= columns)  return;

        pixelsNow = (int)((float)pixelsLeft / (columns - column));

        // handle any remaining columns with evenly divided remaining-width
        while (column < columns - 1) {
            tc = table.getColumnModel().getColumn(column);
            tc.setPreferredWidth(pixelsNow);
            pixelsLeft -= pixelsNow;
            column += 1;
        }
        // last column is leftover; avoid rounding error
        tc = table.getColumnModel().getColumn(column);
        tc.setPreferredWidth(pixelsLeft);
    }
}
