// QEBeanTableModel.java, v1.0, 13 December 2010
/****************************************************
**  QEBeanTableModel
**  Copyright (C) 2010  Rich Franzen
**
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see
**  <http://www.gnu.org/licenses/gpl-3.0.html>.
****************************************************/
package us.r0k.sihwheel;

import java.awt.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import javax.swing.*;
import javax.swing.table.*;
import java.text.DecimalFormat;
import java.util.ArrayList;
import us.r0k.ntc.NTC;
import org.jdesktop.beans.AbstractBean;

// based on design by "Kleopatra", re-thinking my own QEPanel class
// http://forums.oracle.com/forums/thread.jspa?messageID=9200751#9200751
public class QEBeanTableModel extends AbstractTableModel
{
    ArrayList<QEBean> errors = new ArrayList<QEBean>();
    private String columnNames[];
    private PropertyChangeListener errorChangeListener;

    QEBeanTableModel(String columnNames[]) {
	this.columnNames = columnNames;
    }

    /**
     * Returns the column name.
     *
     * @return a name for this column using the string value of the
     * appropriate member in <code>columnNames</code>. If
     * <code>columnNames</code> does not have an entry for this index
     * then the default name provided by the superclass is returned
     */
    @Override
    public String getColumnName(int column)
    {
	String columnName = null;

	if (column < columnNames.length) {
	    columnName = columnNames[column];
	}
	  else
	    columnName = super.getColumnName(column);

	return columnName;
    }

    @Override
    public int getColumnCount()
    {
	if (columnNames == null)  return 0;
        return columnNames.length;
    }

    @Override
    public int getRowCount() {
        return errors.size();
    }

    @Override
    public Class<?> getColumnClass(int columnIndex) {
        switch (columnIndex) {
               // implement as appropriate
        }
        return Object.class;
    }

    @Override
    public Object getValueAt(int rowIndex, int columnIndex)
    {
	DecimalFormat	pct = new DecimalFormat("#0.00%");
	DecimalFormat	f2p = new DecimalFormat("#0.00");
	Object		value = null;

	QEBean error = errors.get(rowIndex);
	if (columnIndex == 0)  return error.getTitle();
	if (error.getCount() <= 0)  return value;
	switch (columnIndex)
	{
	  case 1:
	    int count = error.getCount();
	    if (rowIndex < 3)
		value = count + "/" + error.getEntries() + " error";
	      else
		value = count + " color";
	    if (count > 1)  value = value + "s";
	    break;
	  case 2:
	    value = pct.format(error.getPercentRed());
	    break;
	  case 3:
	    value = pct.format(error.getPercentGreen());
	    break;
	  case 4:
	    value = pct.format(error.getPercentBlue());
	    break;
	  case 5:
	    float avg = (error.getPercentRed() + error.getPercentGreen() +
			error.getPercentBlue()) / 3;
	    value = pct.format(avg);
	    break;
	  case 6:
	    if (error.getEntries() > 0)  break;
	    value = "("  + f2p.format(error.getMinDistance()) +
		    ", " + f2p.format(error.getDistance()) +
		    ", " + f2p.format(error.getMaxDistance()) + ")";
	    break;
        }
        return value;
    }

    public void addError(QEBean error) {
        int oldSize = getRowCount();
        error.addPropertyChangeListener(getPropertyChangeListener());
        errors.add(error);
        fireTableRowsInserted(oldSize, oldSize);
    }

    public QEBean getError(int row) {
        return errors.get(row);
    }

    private PropertyChangeListener getPropertyChangeListener() {
        if (errorChangeListener == null) {
            errorChangeListener = new PropertyChangeListener() {

                @Override
                public void propertyChange(PropertyChangeEvent evt) {
                    int row = errors.indexOf(evt.getSource());
                    fireTableRowsUpdated(row, row);
                }
            };
        }
        return errorChangeListener;
    }

    public void resetAll() {
        for (QEBean error : errors) {
            error.reset();
        }
    }
}

class QEBean extends AbstractBean
{
    private float sumRed, sumGrn, sumBlu;
    private float minDist = 9999999;	// larger than largest min
    private float maxDist = 0;		// smaller than smallest max
    private double sumDist;
    private int count;
    private int entries;
    private String title;

    public QEBean(String title) {
        this.title = title;
    }

    public void reset() {
	sumRed  = 0;
	sumGrn  = 0;
	sumBlu  = 0;
	count   = 0;
	entries = 0;
	sumDist = 0;
	maxDist = 0;
	minDist = 9999999;
	firePropertyChange(null, null, null);
    }

    public void updateData(Color quantized, Color actual)
    {
	float	diff, percent;
	int	not0 = 1;
	if ((quantized == null) || (actual == null)) {
	    System.err.println("updateData(): null color(s)");
	    return;
	}

	diff    = quantized.getRed() - actual.getRed();
	not0    = actual.getRed();
	if (not0 < 8)  not0 = 8;
	sumRed += Math.abs(diff / not0);
	diff    = quantized.getGreen() - actual.getGreen();
	not0    = actual.getGreen();
	if (not0 < 8)  not0 = 8;
	sumGrn += Math.abs(diff / not0);
	diff    = quantized.getBlue() - actual.getBlue();
	not0    = actual.getBlue();
	if (not0 < 8)  not0 = 8;
	sumBlu += Math.abs(diff / not0);
	count  += 1;

	if (entries == 0) {
	    float f = NTC.getDistance();
	    if (f > maxDist)  maxDist = f;
	    if ((f < minDist) && (f > 0))  minDist = f;
	    sumDist += f;
	}

	firePropertyChange(null, null, null);
    }

    public void updateEntries() {
	entries += 1;
	firePropertyChange(null, null, null);
    }

    //------------ getters...
    public String getTitle() {
	return(title);
    }

    public float getPercentRed() {
	if (count == 0)  return 0f;
	return(sumRed / count);
    }
    public float getPercentGreen() {
	if (count == 0)  return 0f;
	return(sumGrn / count);
    }
    public float getPercentBlue() {
	if (count == 0)  return 0f;
	return(sumBlu / count);
    }

    public float getMinDistance() {
	// a young statistics set can appear warped, so override
	float min2 = minDist;
	if (min2 > maxDist)  min2 = maxDist;
	return(min2);
    }
    public float getMaxDistance() {
	return(maxDist);
    }
    public float getDistance() {
	if (count == 0)  return 0f;
	return((float)(sumDist / count));
    }

    public int getCount() {
	return(count);
    }
    public int getEntries() {
	return(entries);
    }
}
