/**
***   SIHwheel.java, v2.60, 14 December 2013
***   by Rich Franzen  (mailto:rich@r0k.us)
***   First Released: 7 September 1998
***
*****************************************************
**  Interactive Color Wheel
**  Copyright (C) 1988 - 2010  Rich Franzen
**
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see
**  <http://www.gnu.org/licenses/gpl-3.0.html>.
*****************************************************
***
***   See bottom of file for revision history.
**/

/**
***   "An interactive test of the Graphics.drawArc routine.
*** Can be run either as a standalone application by
*** typing 'java SIHwheel' or as an applet in the AppletViewer."
***
***   Well, the comment above is for the little applet I started with
*** from the jdk 1.1.6, called 'ArcTest'.
***
***   It has grown into a useful demonstration of my SIH colorspace
*** (described at http://home.att.net/~rocq/png16.html).  Version 1.3
*** is the first one that implements what I really wanted in there at
*** the beginning, although version 1.0 thru 1.2 were posted at
*** earlier times.
***
***   It has further grown (May 2000) into a general interactive color wheel,
*** and gets hundreds of hits a week from people who just want to learn about
*** color, and don't know or care about SIH colorspace.  That's great, and
*** I'm glad people find this generally useful.
**/

/**
***   The goal of this colorspace model is to represent both the
*** advantages and disadvanges of the 16-bit SIH colorspace.
*** Besides, I wanted to see it for myself before designing
*** the png-16 image format (fc16 file type) which can use it.
**/
package us.r0k.sihwheel;

import java.math.BigInteger;
import java.awt.*;
import java.awt.event.*;
import java.io.IOException;
import java.io.File;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import javax.swing.text.html.parser.ParserDelegator;
import javax.swing.text.DefaultEditorKit;
import java.util.Properties;
import java.util.Random;
import us.r0k.ntc.*;

/**
***   Part of the original jdk 'ArcTest' demo was to have an
*** applet that could also be an application.  I have kept that
*** functionality here.
**/
public class SIHwheel extends JApplet
{
    protected static SIHControls controls;
    protected static SIHListPane colorList;
    protected static SIHCanvas paintingCanvas;
    protected static HelpBox logBox;
    protected static Frame ownerFrame;

    // kludge for 1.6.0_22 HTML-in-reused-JVM bug; just make it exist
    private static ParserDelegator workaround;

    // Static initialization of the static NTC data is necessary.
    // Fallback to manual init() if cannot read color name dictionary.
    static { if (!NTC.readCND(null))  NTC.init(); }

    @Override
    public void init()
    { // the static block above should have already executed
        try {
            javax.swing.SwingUtilities.invokeAndWait(new Runnable() {
                public void run() {
                    ownerFrame = getOwnerFrame(getContentPane());
                    JPanel frame = makeContent();
                    setContentPane(frame);
                }
            });
        } catch (Exception e) {
            System.err.println("makeContent() failed to complete: " + e);
            e.printStackTrace();
        }
    }

    public JPanel makeContent()
    {
        workaround = new ParserDelegator();

        // explicitly create logBox with default L&F
        // allows normal copying of rows to the clipboard
        logBox = new HelpBox(ownerFrame, "Color Log",
                             new Dimension(800, 430));

        try {
        UIManager.setLookAndFeel(
            "com.sun.java.swing.plaf.motif.MotifLookAndFeel");
            // Motif L&F does not provide cut/copy/paste automatically
            JTextField field = new JTextField(10);
            InputMap map = field.getInputMap();
            while (map.getParent() != null) {
                map = map.getParent();
            }
            map.put(KeyStroke.getKeyStroke("ctrl C"), DefaultEditorKit.copyAction);
            map.put(KeyStroke.getKeyStroke("ctrl V"), DefaultEditorKit.pasteAction);
            map.put(KeyStroke.getKeyStroke("ctrl A"), DefaultEditorKit.selectAllAction);
            map.put(KeyStroke.getKeyStroke("ctrl X"), DefaultEditorKit.cutAction);
        } catch (UnsupportedLookAndFeelException e) {
            System.err.println("UnsupportedLookAndFeelException: " + e);
        } catch (ClassNotFoundException e) {
            System.err.println("ClassNotFoundException: " + e);
        } catch (InstantiationException e) {
            System.err.println("InstantiationException: " + e);
        } catch (IllegalAccessException e) {
            System.err.println("IllegalAccessException: " + e);
        }

        paintingCanvas = new SIHCanvas(640, 640);
        controls = new SIHControls();
        colorList = new SIHListPane();

        JPanel outer = new JPanel(new GridBagLayout());
        GridBagConstraints gbc = new GridBagConstraints();

        gbc.fill = gbc.BOTH;

        gbc.gridx = 1;
        gbc.gridy = 0;
        outer.add(paintingCanvas, gbc);

        gbc.gridx = 1;
        gbc.gridy = 1;
        outer.add(controls, gbc);

        gbc.gridx = 0;
        gbc.gridy = 0;
        gbc.gridheight = 2;
        outer.add(colorList, gbc);

        return(outer);
    }

    @Override
    public void start()
    {
        controls.setEnabled(true);
        colorList.setEnabled(true);
        logBox.setEnabled(true);

        String firstCND[] = new String[2];
        firstCND[0] = "CND = '" + NTC.getCNDname() + "'";
        firstCND[1] = NTC.names.length + " names";
        logBox.newRow(firstCND);
    }

    @Override
    public void stop()
    { // viewer viewing another tab, browser minimized, etc.
        controls.setEnabled(false);
        colorList.killTimer();
        colorList.setEnabled(false);
    }

    @Override
    public void destroy()
    { // page no longer exists
        colorList.killTimer();
    }

    // returns the parent Frame, JFrame, or PluginEmbeddedFrame
    public static Frame getOwnerFrame(Component component)
    {
        while((component != null) && !(component instanceof Frame)) {
            component = component.getParent();
        }
        return((Frame)component);
    }

    // method expected by programs
    public static void main(String args[])
    {   // JFrame exists only when running as program, not applet
        SwingUtilities.invokeLater(new Runnable()
        {
            @Override
            public void run()
            { // the static block at top should have initialized NTC already
                JFrame f = new JFrame("SIHwheel");
                ownerFrame = f;
                SIHwheel sihWheel = new SIHwheel();

                // in this paradigm, main() does not call init()
                f.setContentPane(sihWheel.makeContent());

                sihWheel.start();

                f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
                f.setResizable(false);
                f.pack();
                f.setLocationRelativeTo(null);  // center on screen
                f.setVisible(true);
            }
        });
    }
}


// Primary, non-list-widget GUI
class SIHControls extends JPanel implements ActionListener // , ChangeListener
{
    private JTextField  intensity;

    private int     tileWidth = 12;
    private int     lastTileWidth = -1;
    private int     myInt = 56;
    private int     lastInt = 56;
    private int     augments = 9;
    private int     lastAugments = 9;
    private boolean rings = true;
    private int     myMaxI = 64;        // given real value a few lines below

    private JToggleButton cBut =
        new JToggleButton("C", true);           // augment cyan
    private JToggleButton bBut =
        new JToggleButton("B", false);          // augment blue
    private JToggleButton mBut =
        new JToggleButton("M", false);          // agument magenta
    private JToggleButton rBut =
        new JToggleButton("R", true);           // augment red
    private JToggleButton gBut =
        new JToggleButton("-G", false);         // narrows green
    private Color bgc    = new Color(20, 20, 20);
    private Color but_fgc = new Color(160, 164, 176);
    private Color acCol = new Color(40, 240, 240);  // augmented foreground colors
    private Color abCol = new Color(80, 80, 255);
    private Color amCol = new Color(240, 40, 240);
    private Color arCol = new Color(240, 40, 40);
    private Color ngCol = new Color(40, 240, 40);

    public SIHControls()
    {
        JButton b  = null;

        this.setForeground(Color.darkGray);
        this.setBackground(bgc);
        this.setLayout(new FlowLayout(FlowLayout.CENTER, 3, 3));
        this.setBorder(BorderFactory.createLineBorder(Color.black, 2));
        myMaxI = SIHwheel.paintingCanvas.maxMaxP1;

        b = new JButton("-i");
        b.setForeground(but_fgc);
        b.setBackground(bgc);
        b.setToolTipText("lower intensity");
        b.addActionListener(this);
        add(b);
        intensity = new JTextField("55", 3);
        intensity.setForeground(Color.white);
        intensity.setBackground(Color.darkGray);
        intensity.setCaretColor(Color.lightGray);
        intensity.setToolTipText("change intensity (0 to " +
            Integer.toString(myMaxI-1) + ")");
        intensity.setActionCommand("iText");
        intensity.addActionListener(this);
        add(intensity);
        b = new JButton("+i");
        b.setForeground(but_fgc);
        b.setBackground(bgc);
        b.setToolTipText("raise intensity");
        b.addActionListener(this);
        add(b);
        add(new JLabel("   "));

        b = new JButton("-");
        b.setForeground(but_fgc);
        b.setBackground(bgc);
        b.setToolTipText("make wheel smaller");
        b.addActionListener(this);
        add(b);
        JToggleButton tb = new JToggleButton("Rings", true);
        tb.setForeground(but_fgc);
        tb.setBackground(bgc);
        tb.setToolTipText("toggle delineating saturation rings");
        tb.addActionListener(this);
        add(tb);
        b = new JButton("+");
        b.setForeground(but_fgc);
        b.setBackground(bgc);
        b.setToolTipText("make wheel bigger");
        b.addActionListener(this);
        add(b);

        add(new JLabel("   "));

        cBut.setForeground(acCol);      // augment cyan (on initially)
        cBut.setBackground(bgc);
        cBut.setToolTipText("toggle 10 or 12 cyan arcs");
        cBut.addActionListener(this);
        add(cBut);
        bBut.setForeground(but_fgc);    // augment blue
        bBut.setBackground(bgc);
        bBut.setToolTipText("toggle 10 or 12 blue arcs");
        bBut.addActionListener(this);
        add(bBut);
        mBut.setForeground(but_fgc);    // augment magenta
        mBut.setBackground(bgc);
        mBut.setToolTipText("toggle 10 or 12 magenta arcs");
        mBut.addActionListener(this);
        add(mBut);
        rBut.setForeground(arCol);      // augment red (on initially)
        rBut.setBackground(bgc);
        rBut.setToolTipText("toggle 10 or 12 red arcs");
        rBut.addActionListener(this);
        add(rBut);
        gBut.setForeground(but_fgc);    // augment red (on initially)
        gBut.setBackground(bgc);
        gBut.setToolTipText("lower/raise visual span of green");
        gBut.addActionListener(this);
        add(gBut);
    }


    public void actionPerformed(ActionEvent ev)
    {
        String label = ev.getActionCommand();

        if (label.equals("+i"))
        {
            myInt += 1;
            if (myInt > myMaxI)  myInt = myMaxI;
            if (!updateIntensity(myInt))  return;
        }
          else if (label.equals("-i"))
          {
            myInt -= 1;
            if (myInt < 0)  myInt = 0;
            if (!updateIntensity(myInt))  return;
          }
          else if (label.equals("Rings"))
            rings = !rings;
          else if (label.equals("+"))
          {
            tileWidth += 1;
            if (tileWidth > 15)  tileWidth = 15;
            if (tileWidth == lastTileWidth)
                return;
              else
                lastTileWidth = tileWidth;
          }
          else if (label.equals("-"))
          {
            tileWidth -= 1;
            if (tileWidth < 10)  tileWidth = 10;
            if (tileWidth == lastTileWidth)
                return;
              else
                lastTileWidth = tileWidth;
          }
          else if (label.equals("C"))
            augments ^= 8;
          else if (label.equals("B"))
            augments ^= 4;
          else if (label.equals("M"))
            augments ^= 2;
          else if (label.equals("R"))
            augments ^= 1;
          else if (label.equals("-G"))
            augments ^= 64;
          else if (label.equals("iText"))
          {
            String intensityS = intensity.getText();
            intensityS.trim();
            if (intensityS.equals("0"))
                myInt = 0;
              else if (intensityS.equals("0.5") ||
                       intensityS.equals("1/2") ||
                       intensityS.equals(".5")  ||
                       intensityS.equals(".50"))
                myInt = 1;
              else
              {
                try
                {
                    myInt = Integer.parseInt(intensityS) + 1;
                }
                  catch (NumberFormatException e)
                  {
                    myInt = lastInt;
                  }
              }
            if (!updateIntensity(myInt))  return;
          }

        if (augments != lastAugments)
        {
            if ((augments & 8) == 8)
                cBut.setForeground(acCol);
              else
                cBut.setForeground(but_fgc);
            if ((augments & 4) == 4)
                bBut.setForeground(abCol);
              else
                bBut.setForeground(but_fgc);
            if ((augments & 2) == 2)
                mBut.setForeground(amCol);
              else
                mBut.setForeground(but_fgc);
            if ((augments & 1) == 1)
                rBut.setForeground(arCol);
              else
                rBut.setForeground(but_fgc);
            if ((augments & 64) == 64)
                gBut.setForeground(ngCol);
              else
                gBut.setForeground(but_fgc);

            lastAugments = augments;
        }

        SIHwheel.paintingCanvas.redraw(myInt, tileWidth, augments, rings);
    }


    /**
    ***   Boolean function updateIntensity() makes it
    *** easy for external methods to update the
    *** intensity component of this class.  It insures
    *** that the provided newIntensity value is legal
    *** (by cropping), and then updates the class fields.
    ***   It returns true if a change actually
    *** occurred, and false otherwise.  It goes thru
    *** the update motions all the time, though.  This
    *** is to avoid data clocking problems and textfield
    *** irregularities.
    ***   As of version 1.40, there are 76 and one-half intensities.
    *** These will be represented linearly from 0 to 76, but 75 is
    *** actually the max possible; intensity 1 corresponds to "one-half".
    **/
    public boolean updateIntensity(int newIntensity)
    {
        boolean         updateWasChange = true;

        if (newIntensity < 0)
            newIntensity = 0;
          else if (newIntensity > myMaxI)
            newIntensity = myMaxI;
        if (newIntensity == lastInt)  updateWasChange = false;

        myInt = newIntensity;
        lastInt = newIntensity;
        if (newIntensity < 2)
        {
            if (newIntensity == 0)
                intensity.setText("0");
              else  // 1 represents one-half
                intensity.setText("0.5");
        }
          else  // 2..76 are mapped to 1..75
            intensity.setText(Integer.toString(newIntensity - 1));

        return(updateWasChange);
    }
}


// list-widget GUI
class SIHListPane extends JPanel
    implements ActionListener, ChangeListener
{
    private String      sortNames[] = { "alphabetically", "by hue", "by hex",
                                        "by luma", "by luminance",
                                        "by average",
                                        "by h:rgb", "by h:lab", "by h:hsb" };
    private String      sortTips[] = {  "alphabetically by name",
                                        "algorithmically by hue",
                                        "numerically by hex color",
                                        "numerically by luma",
                                        "numerically by luminance",
                                        "numerically by average",
        "algorithmically by Hilbert curve on RGB coordinates",
        "algorithmically by Hilbert curve on CIE L*a*b* coordinates",
        "algorithmically by Hilbert curve on HSB coordinates" };
    private int sortCmds[] = {  ColorName.byNAME,  ColorName.byHSI,
                                ColorName.byHEX,   ColorName.byLUMA,
                                ColorName.byLUMI,  ColorName.byAVG,
                                ColorName.byH_RGB, ColorName.byH_LAB,
                                ColorName.byH_HSB };
    private JComboBox           sortBox = new JComboBox(sortNames);

    private JTextField          hexEntry;
    private String              normalHexLabel = new String("Hex Color: ");
    private JLabel              hexLabel = new JLabel(normalHexLabel);
    private JMenu               colorListMenu;
    private JRadioButtonMenuItem colorListItem[] = new JRadioButtonMenuItem[16];
        // use cndFileCount for number of actual entries
    private boolean menuItemsEnabled = false;
    private final String colorListTip[] = { "click color name below to enable Color Lists", "" };


    final int                   ms_MAX = 4000;  // 0.25 frames/sec
    // Slider range of 0 - 4000 will be converted to 2.0 to -2.0,
    // applied as a power of 2.  So physical slider min will be 0
    // but actual minimum delay will be 250 msec.
    private JSlider             msec = new JSlider(JSlider.HORIZONTAL,
                                                   0, ms_MAX, 100);
    String sliderRangeText = new String("(3999 to 250 milliseconds)");
    private JCheckBoxMenuItem   randomSpot;
    private Random              randy = new Random();

    private JList               list;
    final DefaultListModel model = new DefaultListModel();

    private String cndFileList[][] = new String[16][3];
    static final String cfl = new String("/CNDfilelist.properties");
    private int cndFileCount = 0;
    private int reSorts = 0;            // can't access hex widget during first sort
    private int previousSort = 0;       // allow re-use of same sort when CND changed
    private String newCND[] = new String[2];

    private boolean imWorkingHere = false;
    private boolean             hueStats = true;
    final String hsiLabels[] = { " --- Greys and Near-Greys",
                                 " --- Red Sextant",
                                 " --- Yellow Sextant",
                                 " --- Green Sextant",
                                 " --- Cyan Sextant",
                                 " --- Blue Sextant",
                                 " --- Magenta Sextant" };
    private Timer               timer = null;
    private int                 spotsEasterEgg = 0;  // allow overriding the 250ms auto-play limit
    private boolean             spotPlaying = false; // keep track of autoplay mode
    private int                 rawDelay = 0;        // underlying autoplay speed

    private Color bgc = new Color(20, 20, 20);

    public SIHListPane()
    {
        this.setForeground(Color.lightGray);
        this.setBackground(bgc);
        this.setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
        this.setBorder(BorderFactory.createLineBorder(Color.black, 2));

        // create the JList
        // custom renderer courtesy of  pietblok at oracle Java forums
        // http://forums.oracle.com/forums/thread.jspa?threadID=1555347
        // modified for dynamic list
        list = new JList(model);
        getSortedNames(ColorName.byNAME);       // default to alphabetic

        // set some adequate renderer
        list.setCellRenderer(new DefaultListCellRenderer()
        {
            private static final long serialVersionUID = 1L;
            Color fgc;

            // @Override  // not supported in target Java 4 platform
            public Component getListCellRendererComponent(JList list,
                    Object value, int index, boolean isSelected,
                    boolean cellHasFocus) {
                DefaultListCellRenderer renderer = (DefaultListCellRenderer) super
                        .getListCellRendererComponent(list, value, index,
                                isSelected, cellHasFocus);
                renderer.setText(((String[]) value)[1]);
                int color = new BigInteger(((String[]) value)[0], 16)
                        .intValue();
                if (color == -1)
                { // this is a label, not a color
                    renderer.setBackground(Color.white);
                    renderer.setForeground(Color.red);
                }
                  else
                  { // normal color listing
                    Color c = new Color(color);
                    renderer.setBackground(c);
                    fgc = SIHwheel.paintingCanvas.getVisibleForeground(c);
                    renderer.setForeground(fgc);
                  }
                return renderer;
            }
        });

        list.addListSelectionListener(new ListSelectionListener()
        {
            // @Override  // not supported in target Java 4 platform
            public void valueChanged(ListSelectionEvent e)
            {
                // ignore events while rewriting list
                if (imWorkingHere)  return;

                if (!e.getValueIsAdjusting())
                {   // avoid duplicated calls
                    String[] selected = (String[]) list.getSelectedValue();
                    // only redraw() when not a label
                    if (!selected[0].equals("-1"))
                        SIHwheel.paintingCanvas.redraw(selected[0], selected[1]);
                    if (!menuItemsEnabled)
                        setEnabledCNDitems(true);
                }
            }
        });


        if (!readCNDnames())
            JOptionPane.showMessageDialog(null,
                "Unable to build menu 'Color Lists'",
                "SIHwheel Error", JOptionPane.ERROR_MESSAGE);
        ButtonGroup g = new ButtonGroup();
        JMenuBar menuBar = new JMenuBar();
        menuBar.setBorderPainted(false);
        menuBar.setPreferredSize(new Dimension(200, 14));
        colorListMenu = new JMenu("Color Lists");
        colorListMenu.setToolTipText(colorListTip[0]);
        for (int i = 0; i < cndFileCount; i++) {
            final int j = i;  // magic necessary for anonymous actionPerformed
            colorListItem[i] = new JRadioButtonMenuItem(cndFileList[i][1]);
            if (cndFileList[i][2] != null)
                colorListItem[i].setToolTipText(cndFileList[i][2]);
            colorListItem[i].addActionListener( new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    changeCND(j);
                }
            });
            if (cndFileList[i][0].equals(NTC.getCNDfile()))
                colorListItem[i].setSelected(true);
            colorListItem[i].setEnabled(false);
            g.add(colorListItem[i]);
            colorListMenu.add(colorListItem[i]);
        }
        colorListMenu.addSeparator();
        randomSpot = new JCheckBoxMenuItem("Random Spot");
        randomSpot.setToolTipText("When set, Spot jumps randomly, instead of following list");
        colorListMenu.add(randomSpot);
        menuBar.add(colorListMenu);

        menuBar.add(Box.createHorizontalGlue());
        JMenu help = new JMenu("Help");
        JMenuItem sihHelp = new JMenuItem("SIHwheel Help");
        sihHelp.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                Dimension size = new Dimension(880, 600);
                HelpBox hb = new HelpBox(
                    SIHwheel.ownerFrame, "SIHwheel Help", false,
                    "/SIHhelp.html", size);
            }
        });
        help.add(sihHelp);

        JMenuItem log = new JMenuItem("Show Log");
        log.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                SIHwheel.logBox.makeLastRowVisible();
                SIHwheel.logBox.setVisible(true);
            }
        });
        help.add(log);
        help.addSeparator();
        JMenuItem sihHistory = new JMenuItem("SIHwheel Versions");
        sihHistory.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                Dimension size = new Dimension(700, 500);
                HelpBox hb = new HelpBox(
                    SIHwheel.ownerFrame, "SIHwheel History", false,
                    "/SIHhistory.html", size);
            }
        });
        help.add(sihHistory);
        JMenuItem about = new JMenuItem("About");
        about.addActionListener( new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                JOptionPane.showMessageDialog(null,
                    "<html><div align=center>" +
                    "<b>Interactive Color Wheel</b><br>" +
                    "v2.60, 2013-02-16<br>" +
                    "by Rich Franzen, 1998-2013<br>" +
                    "<br>" +
                    "Experiment with all aspects of color,<br>" +
                    "and see names for a multitude of colors.<br>" +
                    "<br>" +
                    "released under the GNU General Public License<br>" +
                    "master URI: href=http://r0k.us/graphics/SIHwheel.html" +
                    "</div></html>",
                    "About", JOptionPane.PLAIN_MESSAGE);
            }
        });
        help.add(about);
        menuBar.add(help);
        add(menuBar);

        JPanel sorts = new JPanel(new FlowLayout(FlowLayout.CENTER));
        sorts.setBorder(BorderFactory.createLineBorder(Color.black, 1));
        sorts.setPreferredSize(new Dimension(200, 20));
        sorts.setBackground(bgc);
        JLabel sortLabel = new JLabel("Sorted");
        sortLabel.setForeground(Color.lightGray);
        sorts.add(sortLabel);
        sortBox.setSelectedIndex(0);
        sortBox.setMaximumRowCount(9);
        sortBox.setToolTipText(sortTips[0]);
        sortBox.setActionCommand("sortby");
        sortBox.addActionListener(this);
        sorts.add(sortBox);
        add(sorts, BorderLayout.LINE_START);

        // standardize how high each cell is
        list.setFixedCellHeight(21);  // defaults to 18 on my system,
                                      // but varies for other folks
        list.setVisibleRowCount(23);  // had been 32 when height was variable
        list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        add(new JScrollPane(list), BorderLayout.LINE_START);

        JPanel spotPanel = new JPanel();
        spotPanel.setBorder(BorderFactory.createLineBorder(Color.black, 1));
        spotPanel.setPreferredSize(new Dimension(200, 45));
        spotPanel.setBackground(bgc);
        JLabel sliderLabel = new JLabel("Spot, the Magic Color Dog",
                                        JLabel.CENTER);
        sliderLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
        sliderLabel.setForeground(Color.lightGray);
        spotPanel.add(sliderLabel);
        msec.setMajorTickSpacing(1000);
        msec.setMinorTickSpacing(250);
        msec.setPaintTicks(true);
        msec.setToolTipText("jump delay " + sliderRangeText);
        msec.addChangeListener(this);
        spotPanel.add(msec);
        add(spotPanel, BorderLayout.LINE_START);

        JPanel hexPanel = new JPanel();
        hexPanel.setBorder(BorderFactory.createLineBorder(Color.black, 1));
        hexPanel.setPreferredSize(new Dimension(200, 20));
        hexPanel.setBackground(bgc);
        hexLabel.setForeground(Color.lightGray);
        hexPanel.add(hexLabel);
        hexEntry = new JTextField("", 6);
        hexEntry.setForeground(Color.white);
        hexEntry.setBackground(Color.darkGray);
        hexEntry.setCaretColor(Color.lightGray);
        hexEntry.setToolTipText("3-digit CSS or 6-digit HTML");
        hexEntry.setActionCommand("hText");
        hexEntry.addActionListener(this);
        hexPanel.add(hexEntry);
        add(hexPanel /* , BorderLayout.PAGE_END */);
    }


    // kludge to prevent super-long CND load times.  If user does not click
    // a color list item, the longest tables can take hours to load.
    private void setEnabledCNDitems(boolean tf)
    {
        for (int i = 0; i < cndFileCount; i++)
            colorListItem[i].setEnabled(tf);
        colorListMenu.setToolTipText(colorListTip[tf ? 1:0]);  // index: int from boolean
        menuItemsEnabled = tf;
    }


    // read the properties file containing names of CND properties files
    private boolean readCNDnames()
    {
        Properties p = NTC.loadProperties(cfl);
        if (p == null)  return(false);

        // each entry has two keys, of form (beginning at index 01):
        ////  CND.file.01 = file name
        ////  CND.name.01 = cnd name
        int i = 1;
        String key, value, index;
        while (i <= cndFileList.length)
        {
            index = Integer.toString(100 + i);
            index = index.substring(1, 3);  // last 2 digits with 0-fill
            key = new String("CND.file." + index);
            value = p.getProperty(key);
            if (value == null)  break;
            // need to precede by "/"; otherwise will look for in CLASSPATH
            cndFileList[i-1][0] = "/" +value;
            key = new String("CND.name." + index);
            value = p.getProperty(key, "no name");
            cndFileList[i-1][1] = value;
            key = new String("CND.tip." + index);
            value = p.getProperty(key);
            cndFileList[i-1][2] = value;  // may be null
            i += 1;
        }
        if (i > cndFileList.length)
            System.err.println("possibly have inaccessible CNDs");
        cndFileCount = i - 1;
        return(cndFileCount > 0);
    }


    // load the supplied Color Name Dictionary and update GUI
    // if null CNDname is null, the default CND is loaded
    private void changeCND(int CNDindex)
    {
        String CNDname = cndFileList[CNDindex][0];
        if (CNDname != null)
            if (CNDname.length() < 1)  CNDname = null;

        if (NTC.readCND(CNDname))
        {   // gather names before getSortedNames() gets fired
            newCND[0] = "CND = '" + NTC.getCNDname() + "'";
            newCND[1] = NTC.names.length + " names";

            // previousSort is index of last selected sort
            sortBox.setSelectedIndex(previousSort);  // auto-fires getSortedNames()

            System.out.println("loaded new CND: " + NTC.getCNDname());
            hueStats = true;    // redisplay for new dictionary
            SIHwheel.logBox.resetQE();

            SIHwheel.logBox.newRow(newCND);
            setEnabledCNDitems(false);          // kludge to prevent extremely long load times
        }
          else
          {
            if (CNDname == null)  CNDname = new String("unknown");
            JOptionPane.showMessageDialog(null,
                "Unable to access color list '" + CNDname + "'",
                "SIHwheel Error", JOptionPane.ERROR_MESSAGE);
          }
    }


    // Sort, and rebuild the GUI list.  sortCmds defined in ColorName.class
    // final static int byHEX   = -1;
    // final static int byNAME  = -2;
    // final static int byHSI   = -3;
    // final static int byLUMI  = -4;
    // final static int byLUMA  = -5;
    // final static int byAVG   = -6;
    // final static int byH_RGB = -7;
    // final static int byH_LAB = -8;
    // final static int byH_HSB = -9;
    private void getSortedNames(int sortCmd)
    {
        ColorName worker;
        String colorname1, colorname2;
        String hexName[];
        String hueReport[] = { "<html>" };

        colorname1 = SIHwheel.paintingCanvas.getLastColorName();

        // perform actual sort
        worker = ColorName.findColorName(sortCmd);

        // clear-out old elements
        imWorkingHere = true;   // disable list event handler
        list.clearSelection();
        model.removeAllElements();
        model.ensureCapacity(NTC.names.length);

        // and re-add new elements
        int lookingFor = 0;     // next hue to look for
        boolean stillNeedRed = true; // both grey and red have hue of 0
        int labelIndex = 0;
        int sextantCount = 0;
        for (int i = 0; i < NTC.names.length; i++)
        {
            hexName = new String[2];
            worker = ColorName.findColorName(i);
            if (sortCmd == ColorName.byHSI)
            { // insert sextant boundary labels
                // 0 <= hue < 60; ten 6-degree sections per sextant
                int hue = worker.getHSI_Hue();
                int sat = worker.getHSI_Sat();
                if ((hue >= lookingFor) ||
                    (stillNeedRed && (sat > 0)))
                {
                    hexName[0] = new String("-1");
                    hexName[1] = hsiLabels[labelIndex];
                    model.addElement(hexName);
                    if (hueStats)
                    {
                        if (labelIndex > 0)
                        {
                            hueReport[0] += sextantCount +
                                hsiLabels[labelIndex - 1] + "<br>";
                            sextantCount = 0;
                        }
                          else
                            hueReport[0] += "-- Hue Breakdown --<br>";
                    }
                    labelIndex +=  1;
                    lookingFor += 10;   // next sextant boundary
                    if (sat > 0)
                    {   // complexity due to red and grey having same hue
                        if (stillNeedRed)  lookingFor -= 10;  // yellow
                        stillNeedRed = false;
                    }
                    hexName = new String[2];    // need fresh copy
                }
            }
            hexName[0] = worker.getHex();
            hexName[1] = worker.getName();
            model.addElement(hexName);
            sextantCount += 1;
        }
        if ((sortCmd == ColorName.byHSI) && hueStats)
        {
            hueReport[0] += sextantCount +
                hsiLabels[labelIndex - 1] + "</html>";
            SIHwheel.logBox.newRow(hueReport);
            hueStats = false;  // only display them one time
        }

        imWorkingHere = false;  // re-enable list event handler
        colorname2 = updateWheelAndList();
        if (colorname2 == null)
        {   // arbitrarily set first list item
            list.setSelectedIndex(0);
            list.ensureIndexIsVisible(0);
        }
        list.requestFocusInWindow();

        if (newCND[0] != null)
        {
            String report = ("Changed '" + colorname1 + "' of old list to '"
                + colorname2 + "' of new."
                + "\nRemember to click color name below"
                + "\nbefore changing Color List again.");
            JOptionPane.showMessageDialog(colorListMenu,
                report, "Color List Changed", JOptionPane.INFORMATION_MESSAGE);
            newCND[0] = null;
        }
        reSorts += 1;
    }

    // Try to get the inner working of list to clear out start-up slowness
    // and match color names between switches of CND.  Returns either the name
    // of the closest found color to the hex in the hexEntry gadget (success)
    // or null (failure).
    private String updateWheelAndList()
    {
        String colname = null;
        String hexS = "000001";
        String h_n[] = {"", ""};

        if (reSorts > 0)
        {   // remember what current hex value was
            hexS = hexEntry.getText();
            NTC.name(hexS, h_n);
// System.out.println(hexS + " --> " + h_n[0] + " = " + h_n[1]);
            updateHex(h_n[0], true);
            SIHwheel.paintingCanvas.redraw(h_n[0].substring(1,7), h_n[1]);
            colname = new String(h_n[1]);
        }
        return colname;
    }


    public void actionPerformed(ActionEvent ev)
    {
        String label = ev.getActionCommand();

        if (label.equals("sortby"))
        {
            int i = list.getSelectedIndex();
            if (i < 0)  i = 0;
            int s = sortBox.getSelectedIndex();
            sortBox.setToolTipText(sortTips[s]);
            getSortedNames(sortCmds[s]);
            previousSort = s;
            if (i >= model.getSize())  i = 0;
            imWorkingHere = true;
            list.setSelectedIndex(i);
            list.requestFocus();
            imWorkingHere = false;
        }
          else if (label.equals("hText"))
          { // only other event is <CR> within hexEntry
            String hexS = hexEntry.getText();
            hexS = hexS.trim();         // delete leading/trailing white space
            if (hexS.equals(""))  return;  // just in case empty

            // back door to override auto-play speed limit
            if (hexS.equals("four"))
            {
                spotsEasterEgg = 0;
                hexEntry.setText("");
                return;
            }
              else if (hexS.equals("nine"))
              {
                spotsEasterEgg = 9;
                seizureWarning();               // possibly resets value to 0
                hexEntry.setText("");
                return;
              }
              else if (hexS.equals("sixteen"))
              {
                spotsEasterEgg = 16;
                seizureWarning();               // possibly resets value to 0
                hexEntry.setText("");
                return;
              }

            hexS = hexS.toUpperCase();  // make all upper case

            // remove possible leading # (technically, removes all #)
            if (hexS.length() > 0)
                if (hexS.charAt(0) == '#')  hexS = hexS.replace("#", "");

            if (NTC.isHex(hexS))
            {   // rewrite clean string and have canvas update itself
                // but first add upport for CSS 3-hex-digit color
                if (hexS.length() == 3)
                {   // abc --> aabbcc
                    char temp[] = {'0', '0', '0', '0', '0', '0' };
                    for (int i=0; i<3; i++)
                    {
                        char twin = hexS.charAt(i);
                        temp[2*i] = twin;
                        temp[2*i+1] = twin;
                    }
                    hexS = new String(temp);
                }
                hexEntry.setText(hexS);
                SIHwheel.paintingCanvas.redraw(hexS, null);
            }
              else
              {
                spotsEasterEgg = 0;     // easily favor no seizure override
                hexEntry.setText("");   // blank the field
              }
          }
          else if (label.equals("tim"))
          { // timer event
            spotJump();
          }
    }


    // handle timer event
    private void spotJump()
    {
        if (randomSpot.isSelected())
        {
            String hexS = getRandomHexColor();
            SIHwheel.paintingCanvas.redraw(hexS, null);
        }
          else
          {
            int listSize = model.getSize();
            int index = list.getSelectedIndex();
            if (index < 0)  index = -1;
            index += 1;
            if (index >= listSize)  index = 0;
            list.setSelectedIndex(index);

            // when possible, keep current name centered
            if (index > 0)  index += 12;  // ensure fully wrap
            if (index >= listSize)  index = listSize - 1;  // and not overbound
            list.ensureIndexIsVisible(index);
          }
    }

    /** Listen to the slider. */
    public void stateChanged(ChangeEvent e)
    {
        JSlider source = (JSlider)e.getSource();

        int fps = (int)source.getValue();
        // smoothing curve courtesy of Aubrey Jaffer
        double fpsF = ms_MAX - fps;             // want slow to left
        fpsF = (fpsF - (ms_MAX / 2)) / 1000;    // 2.0 to -2.0
        int delay = (int)(1000 * Math.pow(2.0, fpsF));
        rawDelay = delay;       // make restoring underlying delay easier

        if (spotsEasterEgg > 4)
        {
            if (delay < 1000)  delay = (delay * 4) / spotsEasterEgg;
              else if (delay > 1000)  delay = (delay * spotsEasterEgg) / (4 * 2);
        }

        if (source.getValueIsAdjusting())
        {
            if (fps == 0)
                hexLabel.setText("STOP");
              else
                hexLabel.setText(delay + " ms");
        }
          else
          {
            hexLabel.setText(normalHexLabel);
            if (fps == 0)
            {
                msec.setToolTipText("jump delay " + sliderRangeText);
                killTimer();
                // update wheel to blank out warnings
                spotJump();
                return;
            }

            if (timer == null)
            {
                timer = new Timer(delay, this);
                timer.setInitialDelay(100);
                timer.setCoalesce(false);
                timer.setActionCommand("tim");
                timer.start();
                spotPlaying = true;
            }
              else
                timer.setDelay(delay);

            msec.setToolTipText(delay + " msec " + sliderRangeText);
            list.requestFocus();  // mark active name again
          }
    }


    // allow applet's stop() and destroy() methods to kill timer
    public void killTimer()
    {
        if (timer != null)
        {
            timer.stop();
            timer = null;
            spotsEasterEgg = 0;
        }
        spotPlaying = false;
    }

    // returns semi-random hex string of form "rrggbb"
    public String getRandomHexColor()
    {
        String  hex;
        int     red, grn, blu, rgb;

        red = randy.nextInt(256);
        if ((red < 12) || (red > 251))
            red = randy.nextInt(256); // reduce, but not eliminate, extremes
        grn = randy.nextInt(256);
        if ((grn < 12) || (grn > 251))
            grn = randy.nextInt(256); // reduce, but not eliminate, extremes
        blu = randy.nextInt(256);
        if ((blu < 12) || (blu > 251))
            blu = randy.nextInt(256); // reduce, but not eliminate, extremes

        // including the extra high-order bit auto-fills in 0s
        rgb = (1 << 24) + (red << 16) + (grn << 8) + blu;
        hex = Integer.toHexString(rgb);
        return(hex.substring(1, 7));
    }

    // give warning when auto-play limits overridden
    private void seizureWarning()
    {
        String report = ("Do you really want to override auto-play seizure limits?");
        int response = JOptionPane.showConfirmDialog(null,
            report, " Seizure Override ??",
            JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE);
        if (response != JOptionPane.YES_OPTION)  spotsEasterEgg = 0;
    }

    // allow external class to know if seizure limits overridden, or set limits back
    public boolean exceedsAutoPlayLimit(boolean limitIt)
    {
        if (!limitIt)
        {
            spotsEasterEgg = 0;
            if (timer != null)
                if (timer.getDelay() > 0)  timer.setDelay(rawDelay);
        }
        return (spotPlaying && (spotsEasterEgg > 0));
    }

    // allow external class to update hexEntry
    public void updateHex(String hex, boolean listToo)
    {
        if (hex == null)  hex = "000000";

        hexEntry.setText(hex);

        if (!listToo)  return;  // leave list alone

        // ensure that we have a 6-character hex string with no leading '#'
        if (hex.length() < 6)  return;
        if (hex.length() == 7)  hex = hex.substring(1, 7);
        if (hex.length() > 6) return;
        if (!NTC.isHex(hex))  return;
        hex = hex.toUpperCase();

        // find the list item with the matching hex value
        String element[];
        int eCount = model.getSize();
        int elIndex = 0;
        for (; elIndex < eCount; elIndex++)
        {
            element = (String[])model.get(elIndex);
            element[0] = element[0].toUpperCase();
            if (hex.equals(element[0]))  break;
        }
        if (elIndex == eCount)
        {
            System.err.println("updateHex() unable to find " + hex);
            return;
        }

        // now that we've found it, set it and scroll so it is visible
        imWorkingHere = true;   // ignore undesired fire of list event
        list.setSelectedIndex(elIndex);

        // Normally, making it visible puts element at top or bottom.
        // Let's try for more of a centering.
        if (elIndex > list.getLastVisibleIndex())
        {
            elIndex += 12;
            if (elIndex >= eCount)  elIndex = eCount - 1;
        }
          else if (elIndex < list.getFirstVisibleIndex())
          {
            elIndex -= 12;
            if (elIndex < 0)  elIndex = 0;
          }
        list.ensureIndexIsVisible(elIndex);
        list.setSelectedIndex(elIndex);
        list.requestFocus();    // ensure outline selected one
        imWorkingHere = false;
        list.requestFocus();    // ensure outline selected two
    }
}


class SIHCanvas extends JPanel
{
    // Class variables that are not primarily associated with wheel or wedge
    private Font        font[];  // 0 is normal, 1 is large, 2 is bold
    private FontMetrics fm[];    // 0 is normal, 1 is large, 2 is bold
    private int         fontH[]; // 0 is normal height; 1 is large, 2 is bold
    private Rectangle   r;       // describes the current canvas
    private int         intensity = 56;         // part of start-up color
    private float       rRat = 1.0f, gRat = 1.0f, bRat = 1.0f;
    private float       wDif = 0.0f;
    final Color         background = new Color(32, 32, 32);
    final Color         foreground = Color.lightGray;
    private Color       l_e[] = { Color.black, Color.white }; // 0 is last, 1 is exact
    private boolean     desiresExact = false;
    private String      hexExact = null;        // string from hex or list widget
    private String      listName = null;        // string from list widget
    private boolean     showQerror = false;  // show quantization error
    private int         resized = 1;         // should be set to 1 on frame resize
    private float       maxI;
    final float         maxMaxI = 74.0f;
    final int           maxMaxP1 = (int)(maxMaxI + 1);
    final double r2d = 180 / Math.PI;
    private String      ntcBuff[] = { "#000000", "unset" };

    // Class variables primarily associated with Color Wheel
    private int         cX, cY;                // center of color wheel
    private Point       lastClick = new Point();
    private Point       lastDots[] = null;
    private Color       lastDotColors[] = null;
    private int         radDelta = 12;  // varies from 10 to 15
    private int         augments = 9;   // cyan and red augmented by default
    private boolean     rings = true;
    private boolean     wheelOnly = false;
    private boolean     lumaOnly = false;
    private boolean     updateDots = true;
    private boolean     exact = false;

    // nerfGr10[] used by getEweUp() and getEweDown() when all hues.
    // also used by colorToSIH()
    final float nerfGrn10[] = { // instead of 10% change each step, use:
         0f / 60f,  9f / 60f, 17f / 60f, 25f / 60f, 33f / 60f,
        40f / 60f, 46f / 60f, 51f / 60f, 55f / 60f, 58f / 60f, 60f / 60f
    };
    // nerfGrn8[] used by getEweUp() and getEweDown() when missing 2 hues
    // (in accordance with pale parables)
    final float nerfGrn8[] = { // instead of 7.5% change each step, use:
         0f / 60f, 11f / 60f, 21f / 60f, 30f / 60f,
        38f / 60f, 45f / 60f, 51f / 60f, 56f / 60f, 60f / 60f
    };

    // target Java 4 platform does not support enum's, so define by hand
    private final int AoI_LUMA_SNAKE = 1;
    private final int AoI_GREY_SNAKE = 2;
    private final int AoI_HUEBORHOOD = 3;
    private final int AoI_BULB_LABEL = 4;
    private final int AoI_LUMA_WEDGE = 5;
    private final int AoI_LAST_COLOR = 6;
    private final int AoI_BRIGHTNESS_BOX = 7;
    private final int AoI_INFO_BOX = 8;
    private final int AoI_VERSION = 9;
    private final int AoI_COLOR_NAME = 10;
    private final int AoI_EXACT_COLOR = 11;
    private final int AoI_LUMA_BUTTON = 12;
    private final int AoI_LUMA_BARS = 13;
    private final int AoI_GUI_INTENSITY = 14;
    private final int AoI_QUANTIZATION_ERROR = 15;
    private final int AoI_COMPLEMENT_BOX = 16;
    private final int AoI_AUTOPLAY_WARNING = 17;
    private final int AoI_AUTOPLAY_OVERRIDE_OFF = 18;

    // Class variables primarily associated with Color Wedge
    private int         wX, wY;                // ULC of color wedge
    private Rectangle   bulb;                   // "thermometer" base of wedge
    private int         wedgeHeight = 5;        // varies between 4 and 6
    final int   wedgeWidth = 35;
    final int   wedgeDot = 7;          // diameter of wedge's intensity dot
    final int   dotToWedge = 12;       // distance of dot center to wedge edge
    private int         wedgeDegrees = 32;                  // part of start-up color
    private float       wedgeDesat = (float)(2.0 / 16.0);   // part of start-up color
    private float       wedgeStep[] = new float[1];
    private boolean     wedgeOnly = false;
    private Color       wedgeColor = Color.black;
    private Color       compColor  = Color.black;
    private Color       lastColor = Color.black;
    private boolean     newClick = false;
    final String myRestore = "[Restore]";               // glue together at rendering
    final String endOverride = " safe autoplay.";       // glue together at rendering
    private int myRestoreW = 20;  // will be adjusted when fonts defined

    // Class variables primarily associated with the satSnake
    private boolean     satSnakeLumaOnly = true;
    private boolean     satSnakeOnly = false;

    // Class variables primarily associated with luma:luminance
    private boolean     lumaBarsOnly = false;
    private int         lumaBarsOnlyMode = 0;   // luma coefficients
    private float       lRed = .299f, lGrn = .587f, lBlu =.114f;  // lumaNosity

    public SIHCanvas(int width, int height)
    {
        Dimension size = new Dimension(width, height);
        this.setPreferredSize(size);

        /**
        ***   This mouse listener is extended from a
        *** skeletal example in _Graphic_Java_, 2nd Edition
        *** by David M. Geary.  (chapter 5, pg 143).
        **/
        addMouseListener(new MouseAdapter()
        {
            public void mouseReleased(MouseEvent e)
            {
                int     x, y;
                int     radius, wheelRadius;
                int     rings;
                Rectangle  llcwh, ulcwh;
                boolean updateWheel = false;

                x = e.getX();
                y = e.getY();
                // System.out.println("(" + x + ", " + y + ")");

                /**
                ***   See if this is a satSnake event (which toggles
                *** the luma mode of the satSnake).  We only care when
                *** the wheel is a color wheel (when wedgeDesat == 1, the
                *** satSnake is invisible).
                **/
                if (wedgeDesat < 1.0)
                {
                    ulcwh = getRectangle(AoI_LUMA_SNAKE);
                    if ((x >= ulcwh.x) && (y >= ulcwh.y))
                        if ((x < ulcwh.x + ulcwh.width) && (y < ulcwh.y + ulcwh.height))
                        {
                            satSnakeLumaOnly = !satSnakeLumaOnly;
                            satSnakeOnly = true;
                            repaint();
                            return;
                        }
                }

                // check for luma mode switch
                // string location is lower left corner
                llcwh = getRectangle(AoI_LUMA_BUTTON);
                if ((x >= llcwh.x) && (x < llcwh.x + llcwh.width) &&
                    (y <= llcwh.y) && (y > llcwh.y - llcwh.height))
                {
                    setLumaNosity(-1);  // advance to next type
                    if (!lumaOnly)  // only repaint grey stuff
                        lumaBarsOnly = true;
                    repaint();
                    return;
                }

                // check for restore button
                if (SIHwheel.colorList.exceedsAutoPlayLimit(true))
                {   // we know autoplay is active and speed-limit overridden
                    llcwh = getRectangle(AoI_AUTOPLAY_OVERRIDE_OFF);
                    llcwh.width = myRestoreW;
                    if ((x >= llcwh.x) && (x < llcwh.x + llcwh.width) &&
                        (y <= llcwh.y) && (y > llcwh.y - llcwh.height))
                    {   // restore normal autoplay speed override
                        SIHwheel.colorList.exceedsAutoPlayLimit(false);
                        repaint();  // schedule getting rid of override messages
                        return;
                    }
                }

                // check for last color block
                ulcwh = getRectangle(AoI_LAST_COLOR);
                if ((x >= ulcwh.x) && (x < ulcwh.x + ulcwh.width) &&
                    (y >= ulcwh.y) && (y < ulcwh.y + ulcwh.height))
                {
                    float  h_s[] = {0f, 0f};
                    intensity = colorToSIH(l_e[0], h_s);
                    wedgeDegrees = (int)(h_s[0] + 0.4999f);
                    wedgeDesat = h_s[1];
                    updateDots = true;
                    repaint();
                    return;
                }

                // check for exact color block
                ulcwh = getRectangle(AoI_EXACT_COLOR);
                if ((x >= ulcwh.x) && (x < ulcwh.x + ulcwh.width) &&
                    (y >= ulcwh.y) && (y < ulcwh.y + ulcwh.height))
                {
                    float  h_s[] = {0f, 0f};
                    desiresExact = true;
                    intensity = colorToSIH(l_e[1], h_s);
                    wedgeDegrees = (int)(h_s[0] + 0.4999f);
                    wedgeDesat = h_s[1];
                    SIHwheel.colorList.updateHex(ntcBuff[0], true);
                    SIHwheel.logBox.updateQE(SIHwheel.logBox.NAME_QE);
                    updateDots = true;
                    repaint();
                    return;
                }

                // check for brightness block
                ulcwh = getRectangle(AoI_BRIGHTNESS_BOX);
                if ((x >= ulcwh.x) && (x < ulcwh.x + ulcwh.width) &&
                    (y >= ulcwh.y) && (y < ulcwh.y + ulcwh.height))
                {
                    System.out.print("Toggling quantization error display");
                    if (showQerror)
                    {
                        showQerror = false;
                        System.out.println(" off.");
                    }
                      else
                      {
                        showQerror = true;
                        System.out.println(" on.");
                      }
                    return;
                }

                // check for complementary block
                ulcwh = getRectangle(AoI_COMPLEMENT_BOX);
                if ((x >= ulcwh.x) && (x < ulcwh.x + ulcwh.width) &&
                    (y >= ulcwh.y) && (y < ulcwh.y + ulcwh.height))
                {
                    float  h_s[] = {0f, 0f};
                    intensity = colorToSIH(compColor, h_s);
                    wedgeDegrees = (int)(h_s[0] + 0.4999f);
                    wedgeDesat = h_s[1];
                    updateDots = true;
                    repaint();
                    return;
                }

                // check for wedge action.  has early returns
                if (x > (wX - 12))
                {   // handle action on wedge portion of window
                    if (x > (wX + 30))
                        return;
                    if ((y < wY) ||
                        (y > (wY + ((maxMaxP1 + 1) * wedgeHeight))))
                        return;

                    rings = maxMaxP1 - ((y - wY) / wedgeHeight);
                    if (rings < 0)  rings = 0;

                    intensity = rings;
                    maxI = calcMaxI(intensity);
                    wheelOnly = true;
                    updateDots = true;
                    repaint();
                    return;
                }

                /**
                ***   We know that the mouse event may be over wheel,
                *** so find out for sure by checking the radius.
                **/
                radius = (int)(Math.sqrt((x-cX)*(x-cX) + (y-cY)*(y-cY)));

                /**
                ***   Calculate a step-function radius
                *** There are usually 16 concentric, saturated rings,
                *** plus 2 inner circles of grey.
                **/
                rings = getRings(intensity);
                wheelRadius = (rings + 2) * radDelta;

                if (radius <= wheelRadius)
                {
                    Point       thisClick = new Point(x, y);

                    // wedgeDesat is based in integral ratio of rings
                    int numerator = (wheelRadius - radius) / radDelta;
                    if (numerator < 0)
                        numerator = 0;
                      else if (numerator > rings)
                        numerator = rings;

                    if (rings == 0)
                    {   // special case intensities (black and 1/2)
                        wedgeDesat = 1.0f;      // assume "grey"
                        if (intensity == 1)     // maps to 0.5
                            if (radius > radDelta)
                                wedgeDesat = 0.0f;      // 1/2 color ring
                    }
                      else
                        wedgeDesat = (float)numerator / rings;

                    x = x - cX;
                    y = cY - y;
                    if (x == 0)
                    {
                        if (y >= 0)
                            wedgeDegrees = 90;
                          else
                            wedgeDegrees = 270;
                    }
                      else
                      {
                        double  radians = Math.atan((double)y / (double)x);
                        if (x > 0)
                        {
                            if (y < 0)  // quad 1
                                wedgeDegrees = (int)(360 + (radians * r2d));
                              else      // quad 4
                                wedgeDegrees = (int)(radians * r2d);
                        }
                          else
                          { // quads 2 and 3
                            wedgeDegrees = (int)(180 + (radians * r2d));
                          }
                      }

                    // determine if this is a double-click
                    if (thisClick.equals(lastClick))
                    {   // handle lumaOnly mode
                        boolean lastLumaOnly = lumaOnly;

                        if (wedgeDesat < 1.0)
                            lumaOnly = false;   // color wheel
                          else
                            lumaOnly = true;    // luma wheel

                        if (lastLumaOnly != lumaOnly)
                            updateWheel = true;
                    }

                    // remember last location for double-click detection
                    lastClick = null;   // hint to garbage collector
                    lastClick = thisClick;

                    // right now -- spaghetti; repaint() sets wedgeOnly to false
                    if (!updateWheel)
                        wedgeOnly = true;
                    updateDots = true;
                    repaint();
                    return;
                }
            }
        });
    }


    /**
    ***   Function getEweUp() calculates the "ewe" factor (corresponding to
    *** mid(r,g,b)) for a given SIH color.  Since I am emulating a quantized
    *** colorspace, there is complexity in the calculations that would not
    *** exist if "everything was float's" or "everything was 24-bit color".
    *** This function is appropriate for the primary hue domains (red, green,
    *** and blue), where the ewe factor increases in the counter-clockwise direction.
    ***   This function returns both ewe and angleDelta[0] (the quantized
    *** hue-wedge width in degrees).
    **/
    private float getEweUp(int domainDegrees, int paleState,
        boolean augmented, boolean skinnyGreen,
        float intensity, float pure60, float desat, float angleDelta[])
    {
        float   ewe;

        switch (paleState)
        {
          case 0:               // full compliment of hues
          default:
            if (augmented)
            {
                ewe = (domainDegrees / 5) * pure60 / 12;
                angleDelta[0] = 5.0f;
            }
              else
              {
                int index = domainDegrees / 6;  //  0 <= index <= 9
                if (skinnyGreen)
                    ewe = pure60 * nerfGrn10[index];
                  else
                    ewe = index * pure60 / 10;
                angleDelta[0] = 6.0f;
              }
            ewe += desat;
            break;

          case 1:               // 2 hues per hue domain sacrificed
            if (augmented)
            {
                ewe = (domainDegrees / 6) * pure60 / 10;
                angleDelta[0] = 6.0f;
            }
              else
              {
                int index = (domainDegrees * 2) / 15;  // 0 <= index <= 7
                if (skinnyGreen)
                    ewe = pure60 * nerfGrn8[index];
                  else
                    ewe = index * pure60 / 8;
                angleDelta[0] = 7.5f;
              }
            ewe += desat;
            break;

          case 2:               // only 2 hues per hue domain
            if (domainDegrees < 30)
                ewe = desat;
              else
                ewe = desat + (intensity - desat) / 2;
            angleDelta[0] = 30.0f;
            break;
        }

        return(ewe);
    }


    /**
    ***   Function getEweDown() calculates the "ewe" factor (corresponding to
    *** mid(r,g,b)) for a given SIH color.  Since I am emulating a quantized
    *** colorspace, there is complexity in the calculations that would not
    *** exist if "everything was float's" or "everything was 24-bit color".
    *** This function is appropriate for the secondary hue domains
    *** (cyan, magenta, and yellow), where the ewe factor decreases
    *** in the counter-clockwise direction.
    ***   This function returns both ewe and angleDelta[0] (the quantized
    *** hue-wedge width in degrees).
    **/
    private float getEweDown(int domainDegrees, int paleState,
        boolean augmented, boolean skinnyGreen,
        float intensity, float pure60, float desat, float angleDelta[])
    {
        float   ewe;

        switch (paleState)
        {
          case 0:               // full compliment of hues
          default:
            if (augmented)
            {
                ewe = ((64 - domainDegrees) / 5) * pure60 / 12;
                angleDelta[0] = 5.0f;
            }
              else
              {
                int index = (65 - domainDegrees) / 6;  // 1 <= index <= 10
                if (skinnyGreen)
                    ewe = pure60 * nerfGrn10[index];
                  else
                    ewe = index * pure60 / 10;
                angleDelta[0] = 6.0f;
              }
            ewe += desat;
            break;

          case 1:               // 2 hues per hue domain sacrificed
            if (augmented)
            {
                ewe = ((65 - domainDegrees) / 6) * pure60 / 10;
                angleDelta[0] = 6.0f;
            }
              else
              {
                int index = (134 - (domainDegrees * 2)) / 15;  // index <= 1 <= 8
                if (skinnyGreen)
                    ewe = pure60 * nerfGrn8[index];
                  else
                    ewe = index * pure60 / 8;
                angleDelta[0] = 7.5f;
              }
            ewe += desat;
            break;

          case 2:               // only 2 hues per hue domain
            if (domainDegrees < 30)
                ewe = intensity;
              else
                ewe = intensity - (intensity - desat) / 2;
            angleDelta[0] = 30.0f;
            break;
        }

        return(ewe);
    }


    /**
    ***   Function getSIHColor() returns the color of the wheel in
    *** accordance with the supplied intensity, desat, and degrees.
    *** Additionally (since it has already done the work for the
    *** calculation anyway, it returns a quantized angleDelta value
    *** representing how wide in degrees the hue is.  According to
    *** the 16-bit SIH rules, this is either 5 or 6.
    ***   Version 1.40 institued the Pale Parables, which mean that the
    *** inner saturation rings sacrifice hue values in order to have an
    *** extra saturation level per intensity.  (The darkest intensities
    *** sacrifice some of their innermost hues in order to gain 5 and
    *** one-half intensity levels.)  So normally the inner ring will have
    *** just 12 hues total, the next outer ring will have the normal
    *** amount minus 12, and the other rings will have the normal amount.
    *** Intensities 1 and 1/2 just have one ring.  1 is fully
    *** saturated and missing 12 hues.  1/2 is fully saturated and contains
    *** just 12 hues.
    **/
    public Color getSIHColor(float intensity, float desat, int degrees,
                             float angleDelta[])
    {
        int     domainDegrees;  // degrees from the base hue (mod 60)
        int     paleState = 0;  // 0 == norm, 1 == norm-12 hues, 2 == 12 hues
        float   pure60;         // amount of purity in a color
        float   red = 0.0f, grn = 0.0f, blu = 0.0f;
        float   pureDesat;
        Color   out;

        /**
        ***   Insure that the input parameters are in a valid range.
        **/
        if (intensity < 0.0)
            intensity = 0.0f;
          else if (intensity > 1.0)
            intensity = 1.0f;

        desat += wDif;          // saturation override
        if (desat < 0.0f)
            desat = 0.0f;
          else if (desat > 1.0)
            desat = 1.0f;
        pureDesat = desat;              // remember for later
        desat = desat * intensity;      // scale it to min(r,g,b)

        // determine if paleState will modify number of hues
        if (intensity == 0.0f)
            paleState = 0;              // doesn't really matter
          else if (intensity <= 3.9/255.0)
            paleState = 2;              // the "one-half" intensity case
          else if (intensity <= 6.79/255.0)
            paleState = 1;              // the "1" case
          else
          {
            int rings = getRings(1 + (int)(intensity * maxMaxI));
            int satRing = (int)((1.0 - pureDesat) * rings + 0.4);

            if (satRing <= 1)
                paleState = 2;          // innermost ring
              else if (satRing == 2)
                paleState = 1;          // next to innermost ring
              else
                paleState = 0;          // normal case (all hues)
          }

        if (degrees < 0)
            degrees = 0;
        degrees = degrees % 360;

        pure60 = intensity - desat;
        domainDegrees = degrees % 60;

        if (degrees < 60)
        {   // red base (to yellow)
            red = intensity;
            blu = desat;
            grn = getEweUp(domainDegrees, paleState, ((augments & 1) == 1),
                false, intensity, pure60, desat, angleDelta);
        }
          else if (degrees < 120)
          { // yellow base (to green)   // never augmented
            grn = intensity;
            blu = desat;
            red = getEweDown(domainDegrees, paleState, false, (augments & 64) == 64,
                intensity, pure60, desat, angleDelta);
          }
          else if (degrees < 180)
          { // green base (to cyan)     // never augmented
            grn = intensity;
            red = desat;
            blu = getEweUp(domainDegrees, paleState, false, (augments & 64) == 64,
                intensity, pure60, desat, angleDelta);
          }
          else if (degrees < 240)
          { // cyan base (to blue)
            blu = intensity;
            red = desat;
            grn = getEweDown(domainDegrees, paleState, ((augments & 8) == 8),
                false, intensity, pure60, desat, angleDelta);
          }
          else if (degrees < 300)
          { // blue base (to magenta)
            blu = intensity;
            grn = desat;
            red = getEweUp(domainDegrees, paleState, ((augments & 4) == 4),
                false, intensity, pure60, desat, angleDelta);
          }
          else
          { // magenta base (to red)
            red = intensity;
            grn = desat;
            blu = getEweDown(domainDegrees, paleState, ((augments & 2) == 2),
                false, intensity, pure60, desat, angleDelta);
          }

        /**
        ***   Apply color blindness factors
        red *= rRat;
        if (red > 1.0)  red = 1.0f;
        grn *= gRat;
        if (grn > 1.0)  grn = 1.0f;
        blu *= bRat;
        if (blu > 1.0)  blu = 1.0f;
        **/

        out = new Color(red, grn, blu);
        return(out);
    }


    /**
    ***   Procedure paintWheel() does the actual work of painting
    *** the color wheel in accordance to the current value of maxI.
    **/
    private void paintWheel(Graphics g)
    {
        int     radius;
        int     myRad;
        float   angleDelta[] = new float[1];
        float   red, grn, blu;
        float   deSat = (float)0.0;
        float   desatDenominator;
        Color   playground;

        cX = (r.width - 60)  / 2;
        cY = r.height - (radDelta * 18) - 16;

        radius = getRings(intensity);
        if (radius == 0)
            desatDenominator = 1.0f;
          else
            desatDenominator = (float)radius;

        radius += 1;    // there are two "grey" rings

        if (wheelOnly)
        {   // paint() hasn't cleared our area, so we'll do it now
            myRad = (18 * radDelta) + 2;
            g.setColor(background);
            g.fillArc(cX - myRad, cY - myRad, myRad * 2, myRad * 2, 0, 360);
        }

        myRad = ((radius + 1) * radDelta) + 1;
        for (; radius > 1; radius--)
        {
            float       startAngle = 0.0f;
            float       frac = 0.0f;
            while (startAngle < 358)
            {
                Color c;
                c = getSIHColor(maxI, deSat, (int)startAngle, angleDelta);

                if (lumaOnly)
                {
                    float       luma;

                    //   I divide by the ?Rat stuff to undo the change
                    // made in getSIHcolor.
                    luma = c.getBlue()  * lBlu / bRat +
                           c.getRed()   * lRed / rRat +
                           c.getGreen() * lGrn / gRat;
                    luma /= 255.0f;     // convert to [0..1] range

                    c = getColorGrey(luma);
                }

                g.setColor(c);
                g.fillArc(cX - myRad, cY - myRad, myRad * 2, myRad * 2,
                    (int)startAngle, (int)(angleDelta[0] + frac));
                startAngle += angleDelta[0];
                frac = startAngle - (int)startAngle;
            }

            if (rings)
            {
                g.setColor(Color.black);
                g.drawArc(cX - myRad, cY - myRad, myRad * 2, myRad * 2, 0, 360);
            }

            deSat += 1 / desatDenominator;
            myRad -= radDelta;
        }

        if (intensity > 0)
        {   // draw triagonals and the "big" filled-in grey circle
            g.setColor(Color.black);
            radius = 18 * radDelta + 2;
            g.drawLine(cX - radius, cY, cX + radius, cY);
            g.drawLine(
                cX - (int)(radius * Math.cos(Math.PI / 3.0)),
                cY - (int)(radius * Math.sin(Math.PI / 3.0)),
                cX + (int)(radius * Math.cos(Math.PI / 3.0)),
                cY + (int)(radius * Math.sin(Math.PI / 3.0)));
            g.drawLine(
                cX - (int)(radius * Math.cos(2.0 * Math.PI / 3.0)),
                cY - (int)(radius * Math.sin(2.0 * Math.PI / 3.0)),
                cX + (int)(radius * Math.cos(2.0 * Math.PI / 3.0)),
                cY + (int)(radius * Math.sin(2.0 * Math.PI / 3.0)));

            playground = getColorGrey(maxI);
            g.setColor(playground);
            g.fillArc(cX - myRad, cY - myRad, myRad * 2, myRad * 2, 0, 360);
            // always draw circle around grey center
            g.setColor(Color.black);
            g.drawArc(cX - myRad, cY - myRad, myRad * 2, myRad * 2, 0, 360);
        }

        myRad -= radDelta;

        // draw "little" filled-in grey circle
        grn = maxI - (2.0f / 255.0f);
        if (grn < 0.0)  grn = 0.0f;
        playground = getColorGrey(grn);
        g.setColor(playground);
        g.fillArc(cX - myRad, cY - myRad, myRad * 2, myRad * 2, 0, 360);
    }


    private void paintWedge(Graphics g)
    {
        int     i;
        int     topY = 50;
        int     numerator, denominator;
        int     myWidth;
        float   stepDegrees[] = new float[1];
        float   deSat;
        float   maxInt;
        Color   c;

        // insure wedgeDegres and wedgeDesat are valid
        if (wedgeDegrees < 0)  wedgeDegrees = 0;
        wedgeDegrees = wedgeDegrees % 360;
        if (wedgeDesat < 0.0)
            wedgeDesat = 0.0f;
          else if (wedgeDesat > 1.0)
            wedgeDesat = 1.0f;

        // wedgeHeight is either 6, 5, or 4.  On smaller sizes,
        // move the wedge a bit closer to center of panel
        wX = r.width - 50 - (6 - wedgeHeight) * 25;
        wY = r.height - 82;

        bulb = null;    /*  keep updated for resizing, etc  */
        bulb = new Rectangle(wX, wY, wedgeWidth + 10, wedgeWidth + 10);

        g.setColor(Color.black);
        g.fillArc(bulb.x, bulb.y, bulb.width, bulb.height, 0, 360);

        wX += 5;
        wY += 8;

        if (lumaOnly || (wedgeDesat == 1.0f))
        {   // draw greyscale wedge
            int         iGrey;
            int         myX;
            boolean     deepGrey = isDeepGrey();
            float       grey;
            float       myMaxLuma = 1.0f;
            int         lines = maxMaxP1 * wedgeHeight;
            float       fLines = (float)lines;

            /**
            ***    If the user has simply selected the inner "dot" of
            *** 0 saturation, then the wedge is a simple cylinder
            *** with intensity ranging from black to full brightness.
            *** However, if in lumaOnly mode, then the maximum
            *** brightness will be scaled to the maximum luminosity
            *** achievable with the current hue and saturation.  Also
            *** the shape of the wedge retains the semi-conical
            *** appearance of color mode.
            **/
            if (lumaOnly && (wedgeDesat != 1.0f))
            {
                c = getSIHColor(1.0f, wedgeDesat, wedgeDegrees, stepDegrees);

                //   Divide by the ?Rat stuff (normall 1.0)
                // to undo any color blindness change made in getSIHcolor.
                myMaxLuma = c.getBlue()  * lBlu / bRat +
                            c.getRed()   * lRed / rRat +
                            c.getGreen() * lGrn / gRat;
                myMaxLuma /= 255.0f;    // convert to [0..1] range
            }

            /**
            ***   Ideally, myWidth and myX would be set in an else clause.
            *** However this confuses the compiler, which in that case
            *** isn't sure that they have values prior to use.
            **/
            myWidth = wedgeWidth;
            myX = wX;

            for (i = 1; i <= lines; i++)
            {
                grey = i / fLines;
                if (lumaOnly)
                {
                    grey *= myMaxLuma;
                    denominator = getRings(((i-1) / wedgeHeight) + 1);
                    myWidth = denominator * 2 + 3;
                    myX = wX + (wedgeWidth - myWidth) / 2;
                }

                /**
                ***   SIH normally has 4096 evenly-spaced levels of grey.
                *** However, when the user has augmented his color pallette,
                *** there are only 256 evenly-spaced levels.  In this case,
                *** though, there are still 60 additional greys at the
                *** darkest levels. this may be thought of as a 16x dynamic
                *** range bonus from 0.0 through 4.0 of the 256-level
                *** pallette.
                **/
                if (deepGrey || (grey <= (64.0 / 4096.0)))
                    c = getColorGrey(grey);
                  else
                  { /*  be anal about quantization  */
                    iGrey = (int)(255.0f * grey);
                    c = new Color(iGrey, iGrey, iGrey);
                  }
                g.setColor(c);
                topY = wY - i;
                g.fillRect(myX, topY, myWidth, 1);
            }
        }
          else
          { // draw color wedge
            for (i = 1; i <= maxMaxP1; i++)
            {
                // calc. num. and den. for the current intensity of _wedge_
                denominator = getRings(i);
                numerator = (int)((denominator * wedgeDesat) + .01);
                myWidth = denominator * 2 + 3;

                maxInt  = calcMaxI(i);

                if (denominator == 0)
                    deSat = 0;
                  else
                    deSat = (float)numerator / denominator;

                c = getSIHColor(maxInt, deSat, wedgeDegrees, stepDegrees);
                g.setColor(c);
                topY = wY - (i * wedgeHeight);
                g.fillRect(wX + (wedgeWidth - myWidth) / 2, topY,
                    myWidth, wedgeHeight);
            }
          }

        wY = topY;      // remember for later mouse events

        // top off wedge with rounded crown
        topY -= wedgeWidth / 3;
        g.fillArc(wX, topY, wedgeWidth, 2 * wedgeWidth / 3, 0, 180);

        labelWedge(g);
        miscLabels(g);
        putDotNearWedge(g);
    }


    private void labelWedge(Graphics g)
    {
        Rectangle ulcwh;
        int     savedX, y;
        int     length;
        int     domainDegrees;
        int     luma;
        int     i, ringCount;
        float   deSat;
        Color   lumaColor, briColor;
        String  myDomain, myLabel;
        String  hexColor;
        StringBuffer  hexWorker;
        float   hsb[] = new float[3];
        float   stepDegrees[] = new float[1];
        final String domains[] =
            { "red", "yellow", "green", "cyan", "blue", "magenta" };

        // base label beneath bulb, but not centered
        ulcwh = getRectangle(AoI_BULB_LABEL);
        savedX = ulcwh.x;       // for use in drawing saturation wedge at bottom

        domainDegrees = wedgeDegrees % 60;
        myDomain = domains[wedgeDegrees / 60];

        /**
        ***   Determining wedgeStep is a bit obtuse; due to the
        *** new ability of augmentation, certain domains may have
        *** 5 degrees per step instead of 6.  I want the user to
        *** see the step-function nature of the degrees, so we'll
        *** byte the bullet and calculate what it really is.
        **/
        wedgeColor = getSIHColor(maxI, wedgeDesat, wedgeDegrees, wedgeStep);
        // System.out.print("wedgeColor = 0x" +
        //     Integer.toHexString(wedgeColor.getRGB()));
        // System.out.println(", wedgeDesat = " + wedgeDesat +
        //     ", wedgeDegrees = " + wedgeDegrees);

        // clear old label
        g.setColor(wedgeColor);
        g.fillRect(ulcwh.x, ulcwh.y, ulcwh.width, ulcwh.height);

        newClick = !lastColor.equals(wedgeColor);

        // write new label
        if (wedgeDesat == 1.0f)
        {   /* deepGrey unless 3 or 4 augments bits are set  */
            if (isDeepGrey())
                myLabel = new String("deep grey");
              else
                myLabel = new String("grey");
        }
          else
          {
            if (wedgeStep[0] <= 0.0)
                stepDegrees[0] = 0.0f;
              else
                stepDegrees[0] = wedgeStep[0] *
                                (int)(domainDegrees / wedgeStep[0]);
            myLabel = new String(myDomain + " + " + stepDegrees[0] + '');
          }

        length = fm[0].stringWidth(myLabel);
        g.setColor(getVisibleForeground(wedgeColor));
        g.drawString(myLabel, ulcwh.x + 5 + ((ulcwh.width - length) / 2),
                     ulcwh.y + 24);

        deSat = (float)((int)(1000 * wedgeDesat) / 10.0);
        myLabel = new String(deSat + "% desat");
        length = fm[0].stringWidth(myLabel);
        g.drawString(myLabel, ulcwh.x + 5 + ((ulcwh.width - length) / 2),
                     ulcwh.y + 24 + 15);

        // re-fill in the portion of the bulb that got overwritten
        g.setColor(Color.black);
        g.fillArc(bulb.x, bulb.y, bulb.width, bulb.height, 180, 180);

        // Place a small luma bar next to color wedge.
        luma = (int)(lRed*wedgeColor.getRed() + lGrn*wedgeColor.getGreen() +
            lBlu*wedgeColor.getBlue() +.49);
        lumaColor = new Color(luma, luma, luma);
        g.setColor(lumaColor);
        ulcwh = getRectangle(AoI_LUMA_WEDGE);
        g.fillRect(ulcwh.x, ulcwh.y, ulcwh.width, ulcwh.height);

        // Place a small last-color square overlapping ULC
        g.setColor(lastColor);
        ulcwh = getRectangle(AoI_LAST_COLOR);
        g.fillRect(ulcwh.x, ulcwh.y, ulcwh.width, ulcwh.height);
        l_e[0] = lastColor;
        lastColor = wedgeColor;

        // Place a brightness box at the LLC of the color wedge.
        // Overlay it just a bit onto the luma bar.
        briColor = new Color(maxI, maxI, maxI);
        g.setColor(briColor);
        ulcwh = getRectangle(AoI_BRIGHTNESS_BOX);
        g.fillRect(ulcwh.x, ulcwh.y, ulcwh.width, ulcwh.height);

        // now show a linear wedge of saturation
        ringCount = getRings(intensity);
        if (ringCount < 1) ringCount = 1;
        savedX += 1;    // the brightess square is a pixel wider
        for (i = 1; i <= ringCount; i++)
        {
            savedX += 7;
            deSat = (ringCount - i) / (float)ringCount;
            briColor = getSIHColor(maxI, deSat, wedgeDegrees, stepDegrees);
            g.setColor(briColor);
            g.fillRect(savedX, r.height - 7, 7, 7);
        }
    }


    // build a 7-digit hex color string of form #rrggbb
    public String colorToHex(Color c)
    {
        int             red, grn, blu;
        String          hexColor;
        StringBuffer    hexWorker;

        red = c.getRed();
        grn = c.getGreen();
        blu = c.getBlue();
        // leave an extra character at high end
        // both forces leading zeroes (if necessary) and leaves room for #
        hexColor = Integer.toHexString(0x1000000 + (red << 16) + (grn << 8) + blu);
        hexWorker = new StringBuffer(hexColor);
        hexWorker.setCharAt(0, '#');
        return(hexWorker.toString());
    }

    /**
    ***   Adds various labeled areas to canvas.
    *** Originally part of labelWedge(), and should be called after it.
    **/
    private void miscLabels(Graphics g)
    {
        Rectangle ulcwh;
        char    my_l;
        int     luma;
        int     hue, bri;
        int     rgb[] = {0, 0, 0};
        float   sat;
        String  hexColor, hexCol2, myLabel;
        float   hsb[] = new float[3];
        Color   exactC, nameForeground;
        String  row[] = new String[6];

        ulcwh = getRectangle(AoI_VERSION);
        g.setColor(foreground);
        g.drawString("SIHwheel, v2.60", ulcwh.x, ulcwh.y);
        g.drawString("rwf, 16.Feb.13", ulcwh.x + 4, ulcwh.y + fontH[0] - 2);
        g.drawString(NTC.names.length + " named colors",
                     ulcwh.x, ulcwh.y + 2 * fontH[0] - 4);

        // display the r:g:b and h:s:b values, scaled appropriately
        // assume worst width for label width (font width varies by platform)
        if (lumaBarsOnlyMode == 0)              // luma
            my_l = 'l';
          else if (lumaBarsOnlyMode == 1)       // luminance
            my_l = 'L';
          else                                  // average
            my_l = 'a';

        ulcwh = getRectangle(AoI_INFO_BOX);
        g.setColor(wedgeColor);
        g.fillRect(ulcwh.x, ulcwh.y, ulcwh.width, ulcwh.height);

        /**
        ***   Several people have requested that I print colors in
        *** HTML hex form.  So I've altered the code to do so.  To
        *** insure it has leading 0's, I start with a number one digit
        *** bigger than it needs to be, then cut off the suffix.  Then I
        *** replace the leading 1 with a '#' so as to look like HTML
        *** color argument.
        **/
        // myLabel = new String("r:g:b," + my_l + " =  "
        //     + red + ": " + grn + ": " + blu + ", " + luma);

        // hexColor also used in the color name field, below
        hexColor = colorToHex(wedgeColor);
        luma = (int)(lRed*wedgeColor.getRed() +
                     lGrn*wedgeColor.getGreen() +
                     lBlu*wedgeColor.getBlue() +.49);
        myLabel = new String("color, " + my_l + " = " + hexColor + ", " + luma);
        g.setColor(getVisibleForeground(wedgeColor));
        g.drawString(myLabel, ulcwh.x + 4, ulcwh.y + 15);
//      if (newClick)  System.out.print(myLabel);

        Color.RGBtoHSB(wedgeColor.getRed(),
                       wedgeColor.getGreen(),
                       wedgeColor.getBlue(), hsb);
        hue = (int)(360 * hsb[0] + .1);
        sat = (float)((int)(1000 * hsb[1]) / 10.0);
        bri = (int)(255 * hsb[2] + .1);
        myLabel = new String("h,s,b =  " + hue + '' + ",  " + sat + "%,  " + bri);
        g.drawString(myLabel, ulcwh.x + 4, ulcwh.y + 30);
//      if (newClick)  System.out.println("  --  " + myLabel);

        // display color complement
        ulcwh = getRectangle(AoI_COMPLEMENT_BOX);
        hsb[0] += 0.5;          // rotate 180 degrees
        if (hsb[0] >= 1.0f)  hsb[0] -= 1.0f;
        compColor = new Color(Color.HSBtoRGB(hsb[0], hsb[1], hsb[2]));
        g.setColor(compColor);
        g.fillRect(ulcwh.x, ulcwh.y, ulcwh.width, ulcwh.height);
        hexCol2 = colorToHex(compColor);
        g.setColor(Color.white);
        g.setColor(getVisibleForeground(compColor));
        myLabel = new String("complement = " + hexCol2);
        g.drawString(myLabel, ulcwh.x + 4, ulcwh.y + 13);

        // clear QError box
        displayQError("z", hexColor, g);

        // display autoplay override messages if override in effect
        if (SIHwheel.colorList.exceedsAutoPlayLimit(true))
        {
            g.setColor(Color.RED);
            ulcwh = getRectangle(AoI_AUTOPLAY_WARNING);
            myLabel = new String("WARNING: Seizure limit for autoplay speed is overridden.");
            g.drawString(myLabel, ulcwh.x + 4, ulcwh.y + fontH[0]*4/5);

            g.setColor(foreground);
            ulcwh = getRectangle(AoI_AUTOPLAY_OVERRIDE_OFF);
            g.setFont(font[2]);         // bold
            g.drawString(myRestore, ulcwh.x, ulcwh.y);
            g.setFont(font[0]);
            g.drawString(endOverride, ulcwh.x + myRestoreW, ulcwh.y);
        }

        // show color name, etc
        if (newClick)  // avoid unnecessary calls to NTC.name()
        {
            exact = NTC.name(hexColor, ntcBuff);
            SIHwheel.logBox.updateQE(SIHwheel.logBox.SEARCH_ERR,
                ntcBuff[0], hexColor);
            // row[0] = ntcBuff[1];  // set later, after prefix added
            row[1] = hexColor;
            row[2] = hexCol2;
            row[5] = ntcBuff[0];
        }
        // quantization error and calculating SIH from rgb can cause mismatches
        if (listName != null)
        {
            if (!listName.equals(ntcBuff[1]))
            {
                row[3] = listName + " (" + hexExact +
                    ") != NTC's " + ntcBuff[1] + " (" + ntcBuff[0] + ")";
//              System.out.println(row[3] + ", " + exact);
                SIHwheel.logBox.updateQE(SIHwheel.logBox.MM_QE,
                    ntcBuff[0], hexExact);
                ntcBuff[1] = listName;
                exact = false;
            }
        }

        if (exact)
        {
            myLabel = new String("=" + ntcBuff[1]);
            if (desiresExact)
                SIHwheel.colorList.updateHex(ntcBuff[0], false);
        }
          else
          {
            if (desiresExact)
            {
                myLabel = new String("@" + ntcBuff[1]);
                SIHwheel.colorList.updateHex(ntcBuff[0], false);
                if (newClick)
                {
                    if (hexExact == null)
                        SIHwheel.logBox.updateQE(SIHwheel.logBox.NAME_QE,
                            hexColor, ntcBuff[0]);
                      else
                        SIHwheel.logBox.updateQE(SIHwheel.logBox.HEX_QE,
                            hexColor, hexExact);  // clicked color name
                }
                row[4] = displayQError(ntcBuff[0], hexColor, g);
            }
              else
              {
                myLabel = new String("~" + ntcBuff[1]);
                if (hexExact != null)
                    SIHwheel.logBox.updateQE(SIHwheel.logBox.HEX_QE,
                        hexColor, hexExact);  // typed in hex value
              }
          }
        row[0] = myLabel;

        ulcwh = getRectangle(AoI_COLOR_NAME);
        g.setColor(wedgeColor);
        g.fillRect(ulcwh.x, ulcwh.y, ulcwh.width, ulcwh.height);
        nameForeground = getVisibleForeground(wedgeColor);
        g.setColor(nameForeground);
        g.setFont(font[1]);     // use large font
        g.drawString(myLabel, ulcwh.x + 8, ulcwh.y + fontH[1]*4/5);
        g.setFont(font[0]);     // and go basck to normal font for everyone else

        // now draw a sqare of the exact color matching name
        ulcwh = getRectangle(AoI_EXACT_COLOR);
        NTC.getRGB(ntcBuff[0], rgb);
        exactC = new Color(rgb[0], rgb[1], rgb[2]);
        l_e[1] = exactC;
        g.setColor(exactC);
        g.fillRect(ulcwh.x, ulcwh.y, ulcwh.width, ulcwh.height);
        g.setColor(nameForeground);
        g.drawRect(ulcwh.x, ulcwh.y, ulcwh.width, ulcwh.height);
        // Note that this ulcwh may also be used by hex entry stuff below

        // Handle update from the hex entry widget.
        if (hexExact == null)
        {
            if (!desiresExact)
                SIHwheel.colorList.updateHex(hexColor, false);
        }
          else
          { // draw a circle of the requested hex color
            // Circle will be to left of the exact color square
            int circX = ulcwh.x - ulcwh.width - 4;
            Color h = new Color(Integer.valueOf(hexExact, 16).intValue());
            g.setColor(h);
            g.fillArc(circX, ulcwh.y, ulcwh.width, ulcwh.width, 0, 360);
            g.setColor(nameForeground);
            g.drawArc(circX, ulcwh.y, ulcwh.width, ulcwh.width, 0, 360);

            // need to pretty up hexExact for displayQError
            hexExact = colorToHex(h);
            if (!desiresExact)
                row[4] = displayQError(hexExact, hexColor, g);

            // show prettified hexExact in hex widget
            SIHwheel.colorList.updateHex(hexExact, false);
            hexExact = null;
          }

        if (newClick)  SIHwheel.logBox.newRow(row);
        listName = null;
        desiresExact = false;
    }


    /**
    ***   Display quantization, both to the console and near the
    *** color name.  If the exact string is the letter "z",
    *** simply clear the QError box and return.
    **/
    private String displayQError(String exact, String actual, Graphics g)
    {
        int     rgbE[] = {0, 0, 0};
        int     rgbA[] = {0, 0, 0};
        int     dR, dG, dB;
        float   bri, pctE;
        String  rpt;

        Rectangle ulcwh = getRectangle(AoI_QUANTIZATION_ERROR);
        if (exact.equals("z"))
        {
            g.setColor(background);
            g.fillRect(ulcwh.x, ulcwh.y, ulcwh.width, ulcwh.height);
            return(null);  // z for ^z, undo
        }

        NTC.getRGB(exact, rgbE);
        NTC.getRGB(actual, rgbA);
        bri = Math.max(rgbE[0], Math.max(rgbE[1], rgbE[2]));
        if (bri < 3)  bri = 3;  // avoid division by 0

        dR = rgbA[0] - rgbE[0];
        dG = rgbA[1] - rgbE[1];
        dB = rgbA[2] - rgbE[2];
        // short circuit if no error
        if ((dR == 0) && (dG == 0) && (dB == 0))  return(null);
        pctE = 100f * (Math.abs(dR) + Math.abs(dG) + Math.abs(dB)) / bri;

        rpt = new String("qe(" + exact + ") = (" +
            dR + "r, " + dG + "g, " + dB + "b) = " + (int)(pctE + 0.4999) + "%");
        // indicate whether error is based on (n)ame color or (h)ex color
        if (hexExact == null)
            rpt = new String("n" + rpt);
          else
            rpt = new String("h" + rpt);

        if (showQerror && (!SIHwheel.colorList.exceedsAutoPlayLimit(true)))
        {   // autoplay message takes priority; it goes in same spot
            g.setColor(foreground);
            g.drawString(rpt, ulcwh.x + 4, ulcwh.y + fontH[0]*4/5);
        }
        return(rpt);
    }


    /**
    ***   Procedure getRectangle() returns (x,y) position of ULC for
    *** the desired rectangle of interest, plus the width and height.
        // enum AoI     // areas of interest
        // {
        //     LUMA_SNAKE, GREY_SNAKE, HUEBORHOOD, BULB_LABEL, LUMA_WEDGE,
        //     LAST_COLOR, BRIGHTNESS_BOX, INFO_BOX, VERSION,
        //     COLOR_NAME, EXACT_COLOR, LUMA_BARS, LUMA_BUTTON, GUI_INTENSITY,
        //     QUANTIZATION_ERROR, COMPLEMENT_BOX
        // }
    **/
    private Rectangle getRectangle(int aoi)
    {
        int     labelWidth = 125;       // label beneath bulb
        int     labelHeight = 50;       // label beneath bulb
        int     labelRightBorder = 4;
        Rectangle rect = new Rectangle();
        Rectangle ulcwh;
        final String myLabel = "r:g:b,L" + " =  255: 255: 255, 25";

        switch (aoi)
        {
          case AoI_HUEBORHOOD:  // hueborhood; 3x3x3 set of close colors
          case AoI_GREY_SNAKE:  // grey snake; mouse active
            rect.x = wX - dotToWedge - wedgeDot - 50;
            rect.y = bulb.y - 32;
            rect.width  = 48;
            rect.height = 48;
            break;

          case AoI_LUMA_SNAKE:  // luma:hue snake; mouse active
            rect.x = wX - dotToWedge - wedgeDot - 58;
            rect.y = wY - 5;
            rect.width  = 56;
            rect.height = 56;
            break;

          case AoI_BULB_LABEL:  // bulb label; beneath bulb, but not centered
            rect.x = r.width  - labelWidth - labelRightBorder;
            rect.y = r.height - labelHeight;
            rect.width  = labelWidth;
            rect.height = labelHeight;
            break;

          case AoI_LUMA_WEDGE:  // small luma wedge to left of label
            rect.x = r.width  - labelWidth  -  4 - labelRightBorder;
            rect.y = r.height - labelHeight + 16;
            rect.width  = 8;
            rect.height = labelHeight - 16;
            break;

          case AoI_LAST_COLOR:  // small last color square; mouse active
            rect.x = r.width  - labelWidth  - 4 - labelRightBorder;
            rect.y = r.height - labelHeight - 4;
            rect.width  = 16;
            rect.height = 16;
            break;

          case AoI_BRIGHTNESS_BOX:      // small brightness box between luma wedge and bulb label
            rect.x = r.width  - labelWidth - labelRightBorder;
            rect.y = r.height - 8;
            rect.width  = 8;
            rect.height = 8;
            break;

          case AoI_INFO_BOX:    // info box above the intensity wedge
            rect.width  = fm[0].stringWidth(myLabel);
            rect.height = 34;
            rect.x = r.width - rect.width - 20;
            rect.y = wY - (wedgeWidth / 3) - 38;
            break;

          case AoI_LUMA_BARS:   // [Luma] bars in LLC
            rect.x = 4;
            rect.y = r.height - 200;
            rect.width  = 10;
            rect.height = 200;
            break;

          case AoI_LUMA_BUTTON: // [Luma] "button"; mouse active
            rect.x = 22;
            rect.y = r.height - 45;
            rect.width  = 44;
            rect.height = 14;
            break;

          case AoI_GUI_INTENSITY:       // "i = " and "tileWidth = " text
            rect.x = 14;
            rect.y = r.height - 40;
            rect.width  = 120;
            rect.height = 40;
            break;

          case AoI_VERSION:     // version and date box to ul of colorwheel
            rect.x = 10;
            rect.y = r.height - (radDelta * 34) - 40;
            if (rect.y < 15)  rect.y = 15;
            // don't need width and height for this case
            break;

          case AoI_COLOR_NAME:  // color name box
            rect.x = 20;
            rect.y = r.height - (radDelta * 34) - fontH[1] * 3 - 4;
            if (rect.y < 10)  rect.y = 10;
            rect.width  = r.width * 4 / 7;
            rect.height = fontH[1] + 8;
            break;

          case AoI_EXACT_COLOR: // exact-color box within name box; mouse active
            ulcwh = getRectangle(AoI_COLOR_NAME);
            rect.width  = fontH[1];
            rect.height = fontH[1];
            rect.x = ulcwh.x + ulcwh.width - rect.width - 4;
            rect.y = ulcwh.y + 4;
            break;

          case AoI_QUANTIZATION_ERROR:  // optional display near COLOR_NAME
          case AoI_AUTOPLAY_WARNING:    // .. or display warning near COLOR_NAME
            ulcwh = getRectangle(AoI_COLOR_NAME);
            rect.width  = 250;
            rect.height = fontH[0] + 4;
            rect.x = ulcwh.x;
            rect.y = ulcwh.y - rect.height;
            break;

          case AoI_COMPLEMENT_BOX:      // display complementary color; mouse active
            ulcwh = getRectangle(AoI_INFO_BOX);
            rect.width  = ulcwh.width;
            rect.height = (ulcwh.height / 2) + 1;
            rect.x = ulcwh.x;
            rect.y = ulcwh.y - rect.height;
            break;

          case AoI_AUTOPLAY_OVERRIDE_OFF:  // pseudo-button to restore seizure override
            rect.width  = myRestoreW + fm[0].stringWidth(endOverride);
            rect.height = fontH[0] + 4;
            rect.x = r.width - rect.width - 10;
            rect.y = rect.height + 2;
            break;
        }

        return(rect);
    }


    /**
    ***   Draw the hueborhood
    **/
    private void drawLumaSnake(Graphics g)
    {
        // lcColor, lcLuma, and lcIndex work together as linked list
        Color   lcColor[] = new Color[29];
        float   lcLuma[]  = new float[29];
        int     lcLink[]  = new int[29];
        int     lcNow = 2;      // first empty slot

        float   angleDelta[] = new float[1];
        int     ctrDegrees;
        float   desat, satInterval;
        float   brightness;

        int     bri, hue, clockHue;
        int     rings;
        int     i, lastI = 0;
        Rectangle       r = new Rectangle();
        RectSnake       rs;
        Color           cc;
        Rectangle       ulcwh;

        // establish first and last elements of list
        lcColor[0] = Color.black;
        lcLuma[0]  = 0.0f;
        lcLink[0]  = 1;
        lcColor[1] = Color.white;
        lcLuma[1]  = 1.1f;
        lcLink[1]  = 0;

        // Determine the 27 colors (including wedgeColor) closest
        // geometrically to wedgeColor.  This is a 3x3x3 array with
        // dimensions of saturation, intensity, and hue.  Store them
        // in the lcColor list by order of luma value.

        // assume colors are 6 degrees apart everywhere
        ctrDegrees = (6 * (wedgeDegrees / 6)) + 3;

        for (bri = intensity - 1; bri <= (intensity + 1); bri += 1)
        {
            rings = getRings(bri);
            if (rings < 1)  rings = 1;
            satInterval = 1.0f / rings;
            brightness = calcMaxI(bri);
            for (hue = ctrDegrees - 6; hue <= (ctrDegrees + 6); hue += 6)
            {
                if (hue < 0)
                    clockHue = hue + 360;
                  else if (hue > 360)
                    clockHue = hue - 360;
                  else
                    clockHue = hue;
                for (desat = wedgeDesat - satInterval;
                     desat <= (wedgeDesat + satInterval + .0001f);
                     desat += satInterval)
                {
                    lcColor[lcNow] = getSIHColor(brightness,
                                     desat, clockHue, angleDelta);
                    lcLuma[lcNow] = (lRed * lcColor[lcNow].getRed() +
                                     lGrn * lcColor[lcNow].getGreen() +
                                     lBlu * lcColor[lcNow].getBlue()) / 255;
                    // now place work color into ordered "linked list"
                    i = 0;      // always links to white
                    while (lcLuma[lcNow] >= lcLuma[i])
                    {
                        lastI = i;
                        i = lcLink[i];
                    }
                    lcLink[lcNow] = i;
                    lcLink[lastI] = lcNow;
                    lcNow += 1;
                }
            }
        }

        // Now we will build the "snake".  It is basically a 6x5 array
        // of color rects, with each rect being 7x8 (for 42x40 pixels).
        // This will be centered within a 48x48 rect of wedgeColor.
        ulcwh = getRectangle(AoI_HUEBORHOOD);
        g.setColor(wedgeColor);
        g.fillRect(ulcwh.x, ulcwh.y, ulcwh.width, ulcwh.height);

        ulcwh.x += (ulcwh.width  - (6 * 7)) / 2;
        ulcwh.y += (ulcwh.height - (5 * 8)) / 2;
        rs = new RectSnake(ulcwh.x, ulcwh.y, 6 * 7, 5 * 8, 7, 8);

        lastI = lcLink[0];
        for (i = 1; i <= 30; i++)
        {
            rs.getNext(r);
            if (i < 28)
            {
                if (lumaOnly)
                    cc = getColorGrey(lcLuma[lastI]);
                  else
                    cc = lcColor[lastI];
                lastI = lcLink[lastI];
                g.setColor(cc);
                g.fillRect(r.x, r.y, r.width, r.height);
            }
              else
              { // show the 3x3 arrays in the 3 empty slots
                int     row, col, x0;

                x0 = r.x + 29 - i;
                lastI = 9*(i - 28) + 2;
                for (row = 0; row < 3; row++)
                    for (col = 0; col < 3; col++)
                    {
                        g.setColor(lcColor[lastI + 3*row + col]);
                        g.fillRect(x0 + 2*col, r.y + 1 + 2*row, 2, 2);
                    }
              }
        }
    }


    /**
    ***   Procedure drawGreySnake() draws a 16x16 array of grey, with
    *** each element being a 3x3 rectangle.
    **/
    private void drawGreySnake(Graphics g)
    {
        int             i;
        float           bri;
        Rectangle       r = new Rectangle();
        RectSnake       rs;
        Color           cc;
        Rectangle       ulcwh;

        ulcwh = getRectangle(AoI_GREY_SNAKE);
        rs = new RectSnake(ulcwh.x, ulcwh.y, ulcwh.width, ulcwh.height, 3, 3);
        for (i = 0; i < 256; i++)
        {
            rs.getNext(r);
            bri = i / 255.0f;
            cc = getColorGrey(bri);
            g.setColor(cc);
            g.fillRect(r.x, r.y, r.width, r.height);
        }
    }


    private void drawSnake(Graphics g)
    {
        if (wedgeDesat == 1.0)
            drawGreySnake(g);
          else
            drawLumaSnake(g);
    }


    /**
    ***   Procedure drawSatSnake() displays an 8x8 rectSnake of the
    *** current saturation ring.  The colors are ordered by luma.
    **/
    private void drawSatSnake(Graphics g)
    {
        // lcColor, lcLuma, and lcIndex work together as linked list
        Color   lcColor[] = new Color[66];
        float   lcLuma[]  = new float[66];
        int     lcLink[]  = new int[66];
        int     lcNow = 2;      // first empty slot

        float   angleDelta[] = new float[1];

        int     hue;
        int     rings;
        int     i, lastI = 0;
        Rectangle       r = new Rectangle();
        RectSnake       rs;
        Color           cc;
        Rectangle       ulcwh;

        ulcwh = getRectangle(AoI_LUMA_SNAKE);

        // clear area if 100% desaturated
        if (wedgeDesat >= 1.0)
        {
            g.setColor(background);
            g.fillRect(ulcwh.x, ulcwh.y, ulcwh.width, ulcwh.height);
            return;
        }

        // establish first and last elements of list
        lcColor[0] = Color.black;
        lcLuma[0]  = 0.0f;
        lcLink[0]  = 1;
        lcColor[1] = Color.white;
        lcLuma[1]  = 1.1f;
        lcLink[1]  = 0;

        // Determine the 64 colors (including wedgeColor) that
        // make up one saturation ring.  To simplify things, never
        // augment blue and magenta (which would make 68 colors).
        hue = 3;
        while (hue < 360)
        {
            lcColor[lcNow] = getSIHColor(maxI,
                             wedgeDesat, hue, angleDelta);
            lcLuma[lcNow] = (lRed * lcColor[lcNow].getRed() +
                             lGrn * lcColor[lcNow].getGreen() +
                             lBlu * lcColor[lcNow].getBlue()) / 255;
            // now place work color into ordered "linked list"
            i = 0;      // always links to white
            while (lcLuma[lcNow] >= lcLuma[i])
            {
                lastI = i;
                i = lcLink[i];
            }
            lcLink[lcNow] = i;
            lcLink[lastI] = lcNow;
            lcNow += 1;

            if (hue < 60)
                hue += 5;               // red
              else if (hue < 180)
                hue += 6;               // yellow and green
              else if (hue < 240)
                hue += 5;               // cyan
              else
                hue += 6;               // blue and magenta
        }

        // Now we will build the "snake".  It is basically a 6x5 array
        // of color rects, with each rect being 7x8 (for 42x40 pixels).
        // This will be centered within a 48x48 rect of wedgeColor.
        rs = new RectSnake(ulcwh.x, ulcwh.y, ulcwh.width, ulcwh.height, 7, 7);

        lastI = lcLink[0];
        for (i = 1; i <= 64; i++)
        {
            rs.getNext(r);
            if (satSnakeLumaOnly)
                cc = getColorGrey(lcLuma[lastI]);
              else
                cc = lcColor[lastI];
            lastI = lcLink[lastI];
            g.setColor(cc);
            g.fillRect(r.x, r.y, r.width, r.height);
        }
    }


    // put current and last 2 dots on color wheel
    private void putDotOnWheel(Graphics g)
    {
        int     numerator, denominator;
        int     radius;
        int     cwX, cwY;       // 'dot' location within wheel
        Color   dotColor;

        // recalculate numerator and denominator for the _wheel_
        denominator = getRings(intensity);
        numerator = (int)((denominator * wedgeDesat) + .01);

        // show where in wheel the color is
        radius  = ((denominator + 2) * radDelta + 1);   // wheelRadius
        radius -= (radDelta * numerator) + 4;           // desaturation dist.

        if (intensity <= 1)
        {   // 0 is black; 1 maps to "1/2" -- which are both special cases
            if (wedgeDesat == 1.0f)
                radius /= 2;            // show grey on innermost ring
        }                               // (usually grey on 2nd innermost)

        cwX = cX + (int)(radius * Math.cos(wedgeDegrees / r2d));
        cwY = cY - (int)(radius * Math.sin(wedgeDegrees / r2d));

        // Place the dot, making sure it is visible.
        // Assumes labelWedge() has been called, defining wedgeColor.
        dotColor = getVisibleForeground(wedgeColor);
        g.setColor(dotColor);
        g.fillArc(cwX - 2, cwY - 2, 5, 5, 0, 360);

        // now show the last three dots, in shrinking size
        if (lastDots == null)
        {   // one-time initialization
            lastDots = new Point[3];
            lastDots[0] = new Point(cwX, cwY);
            lastDots[1] = new Point(cwX, cwY);
            lastDots[2] = new Point(cwX, cwY);
            lastDotColors = new Color[3];
            lastDotColors[0] = Color.white;
            lastDotColors[1] = Color.white;
            lastDotColors[2] = Color.white;
        }
        g.setColor(lastDotColors[0]);
        g.fillRect(lastDots[0].x - 1, lastDots[0].y - 1, 3, 3);
        g.setColor(lastDotColors[1]);
        g.fillRect(lastDots[1].x, lastDots[1].y, 2, 2);
        g.setColor(lastDotColors[2]);
        g.fillRect(lastDots[2].x, lastDots[2].y, 1, 1);

        // and finally, update history arrays
        if (updateDots)
        { // some refreshes do not involve having a new dot
            lastDots[2].x = lastDots[1].x;
            lastDots[2].y = lastDots[1].y;
            lastDotColors[2] = lastDotColors[1];
            lastDots[1].x = lastDots[0].x;
            lastDots[1].y = lastDots[0].y;
            lastDotColors[1] = lastDotColors[0];
            lastDots[0].x = cwX;
            lastDots[0].y = cwY;
            lastDotColors[0] = dotColor;
        }
        updateDots = false;
    }


    /**
    ***   Redraw intensity dot on color wedge.
    **/
    private void putDotNearWedge(Graphics g)
    {
        //  blank the whole "dot column" alongside the wedge
        g.setColor(background);
        g.fillRect(wX - dotToWedge - wedgeDot - 1, wY,
                   wedgeDot * 2 + 2, (maxMaxP1 + 1) * wedgeHeight + 4);

        // and draw the big dot alongside wedge
        g.setColor(foreground);
        g.fillArc(wX - dotToWedge,
            wY + ((maxMaxP1 - intensity) * wedgeHeight) + 3,
            wedgeDot, wedgeDot, 0, 360);
    }


    /**
    ***   Procedure calcMaxI() calculates intensity as a float
    *** between 0.0 and 1.0 inclusive.  Note that as of version
    *** 1.45, there are "75 and one-half" intensities (thanks to the
    *** Pale Parables).  The intensity will be ranged from 0 to 76,
    *** but 75 is actually the max; intensity of 1 is really the
    *** one-half case.  Also, treat 2 special, because it would normally
    *** round to 3/255, but I want it to be 4/255 so that the one-half
    *** is 2/255.
    **/
    private float calcMaxI(int intensity)
    {
        float   maxI;
        final float ratio = 1.0f / maxMaxI;  // maxMaxI is now 74

        // See version history for explanation of Touslah's Pie...
        // final double halfPI = 2.0 / Math.PI;
        // final double ratio  = 1.0 / Math.pow(70.0, halfPI);
        // maxI = Math.pow((double)intensity, halfPI) * ratio;

        if (intensity <= 0)
        {
            maxI = 0.0f;
        }
          else if (intensity == 1)
          { // handle intensity of "one-half"
            maxI = 2.0f / 255;
          }
          else if (intensity == 2)
          { // handle intensity 1 = 4/255 (it would be 3/255 otherwise)
            maxI = 4.0f / 255;
          }
          else if (intensity == 3)
          { // handle intensity 2 = 7/255 (it would by 6/255 otherwise)
            maxI = 7.0f / 255;
          }
          else
          {
            intensity -= 1;  // "1" is the "one-half" case
            if (intensity >= maxMaxI)
            {
                maxI = 1.0f;
            }
              else  // normal case
                maxI = intensity * ratio;
          }

        return(maxI);
    }


    // return white or black based on the supplied background
    public Color getVisibleForeground(Color bg)
    {
        Color fg = Color.white;
        if (bg != null)
        {
            float lumi = bg.getRed()   * .212671f +
                         bg.getGreen() * .715160f +
                         bg.getBlue()  * .072169f;
            if (lumi > 155)  fg = Color.black;
        }
        return(fg);
    }


    /**
    ***   Function getRings() determines how many saturation rings
    *** there are for the supplied intensity.  Because of Shadow Soup
    *** and the Pale Parables, the bottom of the SIH colorspace is
    *** conical, while the upper portion is cylindrical.
    **/
    private int getRings(int intensity)
    {
        int     rings;
        final byte ringArray[] =
        {  0,  0,  1,  2,  2,  3,  3,  4,  4,
           5,  6,  7,  8,  9, 10, 11, 12,
          13, 13, 14, 14, 15, 15, 15, 16 };

        if (intensity < 2)
            rings = 0;          // account for both 0 and 1/2
          else if (intensity < 24)
            rings = (int)ringArray[intensity];
          else
            rings = 16;

        return(rings);
    }


    /**   Boolean function isDeepGrey() determines if the
    *** class 'augments' bits declare that there are 4096
    *** shades of grey.  If not (only 256 shades of grey available)
    *** false is returned.
    **/
    private boolean isDeepGrey()
    {   /* deepGrey unless 3 or 4 augments bits are set  */
        boolean deepGrey = true;

        int aug = augments & 0xf;  // only consider first 4 bits
        if ((aug == 0x7) || (aug == 0xb) ||
            (aug == 0xd) || (aug == 0xe) || (aug == 0xf))
            deepGrey = false;

        return(deepGrey);
    }

    /**
    ***   Function getPseudoGrey() returns a deep-greyscale
    *** value encoded in 24-bit colorspace.  It can uniquely
    *** generate 1,786 levels of grey without degrading from the
    *** color image source.  1786 == (255 * 7) + 1;
    **/
    private Color getPseudoGrey(float grey)
    {
        int     i4k, boost, r, g, b;
        Color   pseudoGrey;

        i4k = (int)(grey * 4095);
        if (i4k >= 4095)
        {   // only one case of "infinity",
            // but the >= also protects against over-bounding (r,g,b)
            r = 255;
            g = 255;
            b = 255;
        }
          else if (i4k >= 0xfe0)
          { // special case near "infinity"
            r = g = b = 0xfe;   // pseudoGrey base (254)
            boost = i4k & 0x1f;
            if (boost >= 28)
            {
                r += 1;
                g += 1;
            }
              else if (boost >= 22)
              {
                g += 1;
                b += 1;
              }
              else if (boost >= 18)
                g += 1;
              else if (boost >= 14)
              {
                r += 1;
                b += 1;
              }
              else if (boost >= 10)
                r += 1;
              else if (boost >= 4)
                b += 1;
          }
          else if (i4k >= 0)
          { // normal case
            r = g = b = i4k >> 4;
            boost = i4k & 0x00f;
            if (boost >= 14)
            {
                r += 1;
                g += 1;
            }
              else if (boost >= 11)
              {
                g += 1;
                b += 1;
              }
              else if (boost >= 9)
                g += 1;
              else if (boost >= 7)
              {
                r += 1;
                b += 1;
              }
              else if (boost >= 5)
                r += 1;
              else if (boost >= 2)
                b += 1;
          }
          else
          { // protect against error condition of negative grey
            r = 0;
            g = 0;
            b = 0;
          }

        pseudoGrey = new Color(r, g, b);
        return(pseudoGrey);
    }

    /**
    ***   Function getColorGrey() returns a new Color value.  If the
    *** primaries are balanced, it will really be grey, but if there
    *** have been color blindness adjustments, then there will be some
    *** tint to it.  The input parameter grey is expected to be bound
    *** between [0..1].
    ***   July 99: added pseudo-grey handling, which extends the
    *** number of greys in a 24-bit colorspace from 256 to
    *** 1786 (7 * 255 + 1).  This is accomplished by allowing
    *** the red, green and blue components to vary by 1, but
    *** only in accordance with a luma-inspired boost.
    **/
    private Color getColorGrey(float grey)
    {
        float   red, grn, blu;
        Color   satGrey;

        /**
        ***   SIH normally has 4096 evenly-spaced levels of grey.
        *** However, when the user has augmented his color pallette,
        *** there are only 256 evenly-spaced levels.  In this case,
        *** though, there are still 60 additional greys at the
        *** darkest levels. this may be thought of as a 16x dynamic
        *** range bonus from 0.0 through 4.0 of the 256-level
        *** pallette.
        **/
        if (isDeepGrey() || (grey <= (64.0 / 4096.0)))
        {   //  We have a PseudoGrey Pixel.
            //  This case does not handle color blindness adjustment.
            satGrey = getPseudoGrey(grey);
        }
          else
          { // Only 256 shades available,
            // but this case handles color-blindness adjustment
            grey += wDif;
            red = grey * rRat;
            if (red > 1.0)  red = 1.0f;
            grn = grey * gRat;
            if (grn > 1.0)  grn = 1.0f;
            blu = grey * bRat;
            if (blu > 1.0)  blu = 1.0f;

            satGrey = new Color(red, grn, blu);
          }

        return(satGrey);
    }


    /**
    ***   Procedure setLumaNosity() sets the luma coefficients to
    *** luma (classic), luminosity (accepted grey weights for modern
    *** monitors), or average (simplistic conversion).  If mode is
    *** passed as -1, the mode is automically advanced to the next type.
    *** The coefficients are from Charles Poyntan's color faq:
    ***  http://www.inforamp.net/~poynton/notes/colour_and_gamma/ColorFAQ.txt
    **/
    private void setLumaNosity(int mode)
    {
        if (mode < 0)
        {
            lumaBarsOnlyMode +=1;
            if (lumaBarsOnlyMode > 2)  lumaBarsOnlyMode = 0;
            mode = lumaBarsOnlyMode;
        }
          else
            lumaBarsOnlyMode = mode;

        if (mode == 0)
        {   // luma weights for color television (NTSC video)
            lRed = .299f;
            lGrn = .587f;
            lBlu = .114f;
        }
          else if (mode == 1)
          { // luminance weights for modern monitors
            lRed = .212671f;
            lGrn = .715160f;
            lBlu = .072169f;
          }
          else
          { // averaged weights (lousy, but many programmers do it!)
            lRed = .3333333f;
            lGrn = .3333333f;
            lBlu = .3333333f;
          }
    }


    /**
    ***   Procedure updateGUIintensity() updates the GUI elements that
    *** have to do with intensity.  As of v1.40, this also includes the
    *** intensity/luma bars on the lower left of canvas.
    **/
    private void updateGUIintensity(Graphics g)
    {
        int     int255;
        Rectangle ulcwh;

        // first blank the text region
        g.setColor(background);
        ulcwh = getRectangle(AoI_GUI_INTENSITY);
        g.fillRect(ulcwh.x, ulcwh.y, ulcwh.width, ulcwh.height);

        // now adjst x and y for the text
        ulcwh.x +=  8;
        ulcwh.y += 10;

        // then write the text
        g.setColor(foreground);
        if (isDeepGrey() & (wedgeDesat == 1.0f))
        {
            int255 = (int)(maxI * 4095 + 0.49f);
            g.drawString("i = " + int255 + " / 4095",
                ulcwh.x, ulcwh.y);
        }
          else
          {
            int255 = (int)(maxI * 255 + 0.49f);
            g.drawString("i = " + int255 + " / 255",
                ulcwh.x, ulcwh.y);
          }
        g.drawString("tileWidth = " + radDelta, ulcwh.x, ulcwh.y + 15);

        // update the textfield intensity gadget
        SIHwheel.controls.updateIntensity(intensity);
    }

    /**
    ***   Procedure updateLumaBars() shows in bar-chart form the
    *** difference between grey intensity and luma level
    *** (NTSC coefficients).
    **/
    private void updateLumaBars(Graphics g)
    {
        int     int255, boldWidth;
        int     lumaHeight;     // height of luma compared to grey int255
        int     lumaPercent;
        int     red = wedgeColor.getRed();
        int     grn = wedgeColor.getGreen();
        int     blu = wedgeColor.getBlue();
        Color   barColor;
        String  boldPart;
        Rectangle ulc;
        double  lumaFrac;
        int     maxLumaHeight = 115;

        // first blank the bar regions, and show the peak value tick mark
        ulc = getRectangle(AoI_LUMA_BARS);
        g.setColor(background);
        g.fillRect(ulc.x, ulc.y, ulc.width, ulc.height);

        int255 = Math.max(Math.max(red, grn), blu);
        if (int255 == 0)  return;  // luminance is not applicable

        // have the grey height vary between 115 and 200
        maxLumaHeight += int255 / 3;

        // draw the intensity and luma bars
        barColor = getColorGrey(int255 / 255.0f);
        g.setColor(barColor);
        g.fillRect(ulc.x, r.height - maxLumaHeight, 5, maxLumaHeight);
        g.setColor(wedgeColor);
        lumaFrac = (red   * lRed + grn * lGrn + blu  * lBlu) / 255;
        lumaHeight = (int)(maxLumaHeight * lumaFrac + 0.4);
        g.fillRect(ulc.x + 5, r.height - lumaHeight, 5, lumaHeight);

        ulc = getRectangle(AoI_LUMA_BUTTON);
        // blank the [luma]= line
        g.setColor(background);
        g.fillRect(ulc.x - 2, ulc.y - 15, 110, 20);
        g.setColor(foreground);
        g.setFont(font[2]);     // bold
        lumaHeight = (int)(lumaFrac * 1000.0 + 0.049);
        lumaPercent = (int)(lumaHeight / 10.0f);
        if (lumaBarsOnlyMode == 0)
        {
            boldPart = new String("[luma]");
            g.drawString(boldPart, ulc.x, ulc.y);
            boldWidth = fm[2].stringWidth(boldPart);
        }
          else if (lumaBarsOnlyMode == 1)
          {
            boldPart = new String("[Lumi]");
            g.drawString(boldPart, ulc.x, ulc.y);
            boldWidth = fm[2].stringWidth(boldPart);
          }
          else
          {
            boldPart = new String("[ave] ");    // need the space at end
            g.drawString(boldPart, ulc.x, ulc.y);
            boldWidth = fm[2].stringWidth(boldPart);
          }
        g.setFont(font[0]);     // back to normal for everyone else
        g.drawString("= " + lumaPercent + "%", ulc.x + boldWidth, ulc.y);
    }


    /**
    ***   Method colorToSIH() translates a given color to the saturation
    *** intensity, and hue values expected by getSIHColor().
    **/
    private int colorToSIH(Color c, float h_s[])
    {
        float   hsbVals[];
        int     intensity;

        // h, s, and b all scaled to 1.0
        hsbVals = Color.RGBtoHSB(c.getRed(), c.getGreen(), c.getBlue(), null);
        h_s[0] = (float)(hsbVals[0] * 2 * Math.PI * r2d);  // degrees

        // if we're nerfing green, we need to distort the degrees
        // this will make more yellows and cyans, and less greeens
        if ((augments & 64) == 64)
        {   // The logic here should utilize that of nerfGrn10[].
            // Uses the inverse distortion seen in getEweUp() and getEweDown().
            if ((h_s[0] >= 60f) && (h_s[0] < 120f))
            {   // can't figure a clean way to put this in a loop
                if ((h_s[0] >= 60f + (1f - nerfGrn10[10]) * 60f) &&
                    (h_s[0] <  60f + (1f - nerfGrn10[ 9]) * 60f))
                    h_s[0] = 63f;
                  else if ((h_s[0] >= 60f + (1f - nerfGrn10[9]) * 60f) &&
                           (h_s[0] <  60f + (1f - nerfGrn10[8]) * 60f))
                    h_s[0] = 69f;
                  else if ((h_s[0] >= 60f + (1f - nerfGrn10[8]) * 60f) &&
                           (h_s[0] <  60f + (1f - nerfGrn10[7]) * 60f))
                    h_s[0] = 75f;
                  else if ((h_s[0] >= 60f + (1f - nerfGrn10[7]) * 60f) &&
                           (h_s[0] <  60f + (1f - nerfGrn10[6]) * 60f))
                    h_s[0] = 81f;
                  else if ((h_s[0] >= 60f + (1f - nerfGrn10[6]) * 60f) &&
                           (h_s[0] <  60f + (1f - nerfGrn10[5]) * 60f))
                    h_s[0] = 87f;
                  else if ((h_s[0] >= 60f + (1f - nerfGrn10[5]) * 60f) &&
                           (h_s[0] <  60f + (1f - nerfGrn10[4]) * 60f))
                    h_s[0] = 93f;
                  else if ((h_s[0] >= 60f + (1f - nerfGrn10[4]) * 60f) &&
                           (h_s[0] <  60f + (1f - nerfGrn10[3]) * 60f))
                    h_s[0] = 99f;
                  else if ((h_s[0] >= 60f + (1f - nerfGrn10[3]) * 60f) &&
                           (h_s[0] <  60f + (1f - nerfGrn10[2]) * 60f))
                    h_s[0] = 105f;
                  else if ((h_s[0] >= 60f + (1f - nerfGrn10[2]) * 60f) &&
                           (h_s[0] <  60f + (1f - nerfGrn10[1]) * 60f))
                    h_s[0] = 111f;
                  else
                    h_s[0] = 117f;
            }
            if ((h_s[0] >= 120f) && (h_s[0] < 180f))
            {   // can't figure a clean way to put this in a loop
                if ((h_s[0] >= 120f + nerfGrn10[0] * 60f) &&
                    (h_s[0] <  120f + nerfGrn10[1] * 60f))
                    h_s[0] = 123f;
                  else if ((h_s[0] >= 120f + nerfGrn10[1] * 60f) &&
                           (h_s[0] <  120f + nerfGrn10[2] * 60f))
                    h_s[0] = 129f;
                  else if ((h_s[0] >= 120f + nerfGrn10[2] * 60f) &&
                           (h_s[0] <  120f + nerfGrn10[3] * 60f))
                    h_s[0] = 135f;
                  else if ((h_s[0] >= 120f + nerfGrn10[3] * 60f) &&
                           (h_s[0] <  120f + nerfGrn10[4] * 60f))
                    h_s[0] = 141f;
                  else if ((h_s[0] >= 120f + nerfGrn10[4] * 60f) &&
                           (h_s[0] <  120f + nerfGrn10[5] * 60f))
                    h_s[0] = 147f;
                  else if ((h_s[0] >= 120f + nerfGrn10[5] * 60f) &&
                           (h_s[0] <  120f + nerfGrn10[6] * 60f))
                    h_s[0] = 153f;
                  else if ((h_s[0] >= 120f + nerfGrn10[6] * 60f) &&
                           (h_s[0] <  120f + nerfGrn10[7] * 60f))
                    h_s[0] = 159f;
                  else if ((h_s[0] >= 120f + nerfGrn10[7] * 60f) &&
                           (h_s[0] <  120f + nerfGrn10[8] * 60f))
                    h_s[0] = 165f;
                  else if ((h_s[0] >= 120f + nerfGrn10[8] * 60f) &&
                           (h_s[0] <  120f + nerfGrn10[9] * 60f))
                    h_s[0] = 171f;
                  else
                    h_s[0] = 177f;
            }
        }

        h_s[1] = 1.0f - hsbVals[1];     // our mechanics use "desat"
        intensity = (int)(hsbVals[2] * maxMaxP1 + 0.9999f);
        return(intensity);
    }


    /**
    ***   Procedure defineFont() defines the font and font metrics for
    *** use in this applet/program.
    **/
    private void defineFont(Graphics g)
    {
        font  = new Font [3];
        fm    = new FontMetrics [3];
        fontH = new int [3];

        font[0] = new Font("SansSerif", Font.PLAIN, 14);
        font[1] = new Font("SansSerif", Font.PLAIN, 24);
        font[2] = new Font("Monospaced", Font.BOLD, 14);
        g.setFont(font[0]);
        fm[0] = g.getFontMetrics(font[0]);
        fm[1] = g.getFontMetrics(font[1]);
        fm[2] = g.getFontMetrics(font[2]);
        fontH[0] = fm[0].getHeight();
        fontH[1] = fm[1].getHeight();
        fontH[2] = fm[2].getHeight();

        myRestoreW = fm[2].stringWidth(myRestore);
    }


    // was named paint() in awt version
    public void paintComponent(Graphics g)
    {
        Rectangle ulc;

        if (font == null)  defineFont(g);

        if (r != null)
            r = null;   /*  re-init to handle resizing of frame  */
        r = getBounds();

        // for clean double-buffering
        super.paintComponent(g);
        g.setColor(background);
        g.fillRect(0, 0, r.width, r.height);

        maxI = calcMaxI(intensity);

        // delineate the pane division with a black line
        g.setColor(Color.black);
        g.fillRect(0, r.height - 2, r.width, 2);
        r.height -= 2;          // and lie about the height from now on

        paintWedge(g);
        drawSnake(g);
        drawSatSnake(g);
        paintWheel(g);
        putDotOnWheel(g);

        updateGUIintensity(g);
        updateLumaBars(g);
    }


    // Allow external class to force changes.
    // Caller should pre-validate the arguments.
    public void redraw(int intensity, int tileWidth, int augments,
                       boolean rings)
    {
        if (tileWidth <= 10)
            wedgeHeight = 4;
          else if (tileWidth <= 12)
            wedgeHeight = 5;
          else
            wedgeHeight = 6;
        this.intensity = intensity;
        this.radDelta = tileWidth;
        this.augments = augments;
        if (rings == this.rings)
            updateDots = true;  // no need to update if rings went on or off
          else
            this.rings = rings;
        repaint();
    }

    // Allow external class to force changes.
    // Caller should pre-validate that hexS is valid hex string
    public void redraw(String hexS, String name)
    {
        float  h_s[] = {0f, 0f};
        Color  c = new Color(Integer.valueOf(hexS, 16).intValue());
        intensity = colorToSIH(c, h_s);
        wedgeDegrees = (int)(h_s[0] + 0.4999f);
        wedgeDesat = h_s[1];
        hexExact = hexS;
        if (name != null)
        {
            desiresExact = true;
            listName = name;
            SIHwheel.logBox.updateQE(SIHwheel.logBox.MM_QE);
        }
        updateDots = true;
        SIHwheel.logBox.updateQE(SIHwheel.logBox.HEX_QE);
        repaint();
    }

    // Allow external class to get last color name
    public String getLastColorName()
    {
        return ntcBuff[1];
    }
}


/**
***   Class RectSnake forms a rectangular, spiraling-inward array
*** of locations.  It starts at the upper left of the defining
*** location.  (It would not be too hard to make it start at other
*** locations (probably based on sign of incX and incY), but for the
*** purposes of SIHwheel, ULC is fine.)
**/
class RectSnake extends Rectangle
{
    // int x, y, width, height; // inherited from Rectangle
    private int direction;      // 0 = right, 1 = down, 2 = left, 3 = up
    private int incX, incY;
    private int elementCount;
    private Rectangle   now = new Rectangle();
    private Point       p = new Point();

    RectSnake(int x, int y, int width, int height, int incX, int incY)
    {
        // validate the parameters
        if (width < 1)   width  = 1;
        if (height < 1)  height = 1;
        if (incX < 1)  incX = 1;
        if (incY < 1)  incY = 1;
        if (incX > width)   incX = width;
        if (incY > height)  incY = height;

        this.x = x;
        this.y = y;
        this.width  = width;
        this.height = height;
        this.incX = incX;
        this.incY = incY;

        reset();
    }

    public void reset()
    {
        now.setBounds(x, y, width, height);
        elementCount = (width / incX) * (height / incY);
        p.x = x - incX;
        p.y = y;
        direction = 0;
    }


    /**
    ***   The getNext() procedure is the primary method used by this
    *** class.  It alters r, filling it with the values for the next
    *** element of the RectSnake.  I expect that the rect r will be
    *** supplied from the outside, but if a null is passed, then
    *** a new rect is made.
    **/
    public void getNext(Rectangle r)
    {
        if (elementCount <= 0)  reset();
        if (r == null)  r = new Rectangle();

        switch (direction)
        {
          case 0:       // right
            p.x += incX;
            if (p.x >= (now.x + now.width))
            {
                p.x -= incX;
                direction = 1;
                p.y += incY;
            }
            break;

          case 1:       // down
            p.y += incY;
            if (p.y >= (now.y + now.height))
            {
                p.y -= incY;
                direction = 2;
                p.x -= incX;
            }
            break;

          case 2:       // left
            p.x -= incX;
            if (p.x < now.x)
            {
                p.x += incX;
                direction = 3;
                p.y -= incY;
                now.y += incY;
            }
            break;

          case 3:       // up
            p.y -= incY;
            if (p.y < now.y)
            {
                p.y += incY;
                direction = 0;
                p.x += incX;
                now.x += incX;
                now.width -= incX * 2;
                now.height -= incY * 2;
            }
            break;
        }

        r.setBounds(p.x, p.y, incX, incY);
        elementCount -= 1;
    }
}

// Version History:  moved to SIHhistory.html
